from typing import Union

from py_pdf_term.pdftoxml import PDFnXMLElement

from ...configs import XMLLayerConfig
from .base import BaseXMLLayerCache


class XMLLayerNoCache(BaseXMLLayerCache):
    def __init__(self, cache_dir: str) -> None:
        pass

    def load(
        self, pdf_path: str, config: XMLLayerConfig
    ) -> Union[PDFnXMLElement, None]:
        pass

    def store(self, pdfnxml: PDFnXMLElement, config: XMLLayerConfig) -> None:
        pass

    def remove(self, pdf_path: str, config: XMLLayerConfig) -> None:
        pass



from typing import Dict, List, Optional, Type

from py_pdf_term._common.data import ScoredTerm
from py_pdf_term.candidates import (
    DomainCandidateTermList,
    PageCandidateTermList,
    PDFCandidateTermList,
)

from .data import DomainStylingScoreList, PageStylingScoreList, PDFStylingScoreList
from .scores import BaseStylingScore, ColorScore, FontsizeScore


class StylingScorer:
    def __init__(
        self, styling_score_clses: Optional[List[Type[BaseStylingScore]]] = None
    ) -> None:
        if styling_score_clses is None:
            styling_score_clses = [FontsizeScore, ColorScore]

        self._styling_score_clses = styling_score_clses

    def score_domain_candidates(
        self, domain_candidates: DomainCandidateTermList
    ) -> DomainStylingScoreList:
        return DomainStylingScoreList(
            domain_candidates.domain,
            list(map(self.score_pdf_candidates, domain_candidates.pdfs)),
        )

    def score_pdf_candidates(
        self, pdf_candidates: PDFCandidateTermList
    ) -> PDFStylingScoreList:
        return PDFStylingScoreList(
            pdf_candidates.pdf_path,
            list(map(self._score_page_candidates, pdf_candidates.pages)),
        )

    def _score_page_candidates(
        self, page_candidates: PageCandidateTermList
    ) -> PageStylingScoreList:
        styling_scores: Dict[str, float] = {
            candidate.lemma(): 1.0 for candidate in page_candidates.candidates
        }

        for styling_score_cls in self._styling_score_clses:
            styling_score = styling_score_cls(page_candidates)

            scores: Dict[str, float] = dict()
            for candidate in page_candidates.candidates:
                candidate_lemma = candidate.lemma()
                score = styling_score.calculate_score(candidate)
                if candidate_lemma not in scores or score > scores[candidate_lemma]:
                    scores[candidate_lemma] = score

            for candidate_lemma in styling_scores:
                styling_scores[candidate_lemma] *= scores[candidate_lemma]

        ranking = list(map(lambda item: ScoredTerm(*item), styling_scores.items()))
        ranking.sort(key=lambda scored_term: -scored_term.score)
        return PageStylingScoreList(page_candidates.page_num, ranking)



from .data import DomainCandidateTermList, PageCandidateTermList, PDFCandidateTermList
from .extractor import CandidateTermExtractor

# isort: unique-list
__all__ = [
    "CandidateTermExtractor",
    "DomainCandidateTermList",
    "PDFCandidateTermList",
    "PageCandidateTermList",
]



from io import BytesIO
from typing import BinaryIO, Callable, Optional
from xml.etree.ElementTree import fromstring

from pdfminer.layout import LAParams
from pdfminer.pdfinterp import PDFPageInterpreter, PDFResourceManager
from pdfminer.pdfpage import PDFPage

from .data import PDFnXMLElement, PDFnXMLPath
from .textful import TextfulXMLConverter


class PDFtoXMLConverter:
    def __init__(self, open_bin: Callable[[str, str], BinaryIO] = open):  # type: ignore
        self.open_bin = open_bin

    def convert_as_file(
        self,
        pdf_path: str,
        xml_path: str,
        nfc_norm: bool = True,
        include_pattern: Optional[str] = None,
        exclude_pattern: Optional[str] = None,
    ) -> PDFnXMLPath:
        pdf_io, xml_io = self.open_bin(pdf_path, "rb"), self.open_bin(xml_path, "wb")
        self._run(pdf_io, xml_io, nfc_norm, include_pattern, exclude_pattern)
        pdf_io.close(), xml_io.close()

        return PDFnXMLPath(pdf_path, xml_path)

    def convert_as_element(
        self,
        pdf_path: str,
        nfc_norm: bool = True,
        include_pattern: Optional[str] = None,
        exclude_pattern: Optional[str] = None,
    ) -> PDFnXMLElement:
        pdf_io, xml_io = self.open_bin(pdf_path, "rb"), BytesIO()
        self._run(pdf_io, xml_io, nfc_norm, include_pattern, exclude_pattern)
        xml_element = fromstring(xml_io.getvalue().decode("utf-8"))
        pdf_io.close(), xml_io.close()
        return PDFnXMLElement(pdf_path, xml_element)

    def _run(
        self,
        pdf_io: BinaryIO,
        xml_io: BinaryIO,
        nfc_norm: bool,
        include_pattern: Optional[str],
        exclude_pattern: Optional[str],
    ) -> None:
        manager = PDFResourceManager()
        laparams = LAParams(char_margin=2.0, line_margin=0.5, word_margin=0.2)
        converter = TextfulXMLConverter(
            manager,
            xml_io,
            laparams=laparams,
            nfc_norm=nfc_norm,
            include_pattern=include_pattern,
            exclude_pattern=exclude_pattern,
        )
        page_interpreter = PDFPageInterpreter(manager, converter)

        pages = PDFPage.get_pages(pdf_io)  # type: ignore
        converter.write_header()
        for page in pages:
            page_interpreter.process_page(page)  # type: ignore
        converter.write_footer()



from py_pdf_term._common.data import Term
from py_pdf_term.tokenizer import Token
from ....classifiers import EnglishTokenClassifier

from ..base import BaseEnglishCandidateTermFilter


class EnglishNumericFilter(BaseEnglishCandidateTermFilter):
    def __init__(self) -> None:
        self._classifier = EnglishTokenClassifier()

    def is_candidate(self, scoped_term: Term) -> bool:
        return not self._is_numeric_phrase(scoped_term)

    def _is_numeric_phrase(self, scoped_term: Term) -> bool:
        def is_numeric_or_meaningless(token: Token) -> bool:
            return token.pos == "NUM" or self._classifier.is_meaningless(token)

        return all(map(is_numeric_or_meaningless, scoped_term.tokens))



from typing import Optional

from py_pdf_term.stylings import PDFStylingScoreList, StylingScorer

from ..caches import DEFAULT_CACHE_DIR
from ..configs import StylingLayerConfig
from ..mappers import StylingLayerCacheMapper, StylingScoreMapper
from .candidate import CandidateLayer


class StylingLayer:
    def __init__(
        self,
        candidate_layer: CandidateLayer,
        config: Optional[StylingLayerConfig] = None,
        styling_score_mapper: Optional[StylingScoreMapper] = None,
        cache_mapper: Optional[StylingLayerCacheMapper] = None,
        cache_dir: str = DEFAULT_CACHE_DIR,
    ) -> None:
        if config is None:
            config = StylingLayerConfig()
        if styling_score_mapper is None:
            styling_score_mapper = StylingScoreMapper.default_mapper()
        if cache_mapper is None:
            cache_mapper = StylingLayerCacheMapper.default_mapper()

        styling_score_clses = styling_score_mapper.bulk_find(config.styling_scores)
        cache_cls = cache_mapper.find(config.cache)

        self._scorer = StylingScorer(styling_score_clses=styling_score_clses)
        self._cache = cache_cls(cache_dir=cache_dir)
        self._config = config

        self._candidate_layer = candidate_layer

    def create_pdf_styling_scores(self, pdf_path: str) -> PDFStylingScoreList:
        styling_scores = self._cache.load(pdf_path, self._config)

        if styling_scores is None:
            pdf_candidates = self._candidate_layer.create_pdf_candidates(pdf_path)
            styling_scores = self._scorer.score_pdf_candidates(pdf_candidates)

        self._cache.store(styling_scores, self._config)

        return styling_scores

    def remove_cache(self, pdf_path: str) -> None:
        self._cache.remove(pdf_path, self._config)



from typing import Type

from ...caches import BaseXMLLayerCache, XMLLayerFileCache, XMLLayerNoCache
from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class XMLLayerCacheMapper(BaseMapper[Type[BaseXMLLayerCache]]):
    @classmethod
    def default_mapper(cls) -> "XMLLayerCacheMapper":
        default_mapper = cls()

        cache_clses = [XMLLayerNoCache, XMLLayerFileCache]
        for cache_cls in cache_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{cache_cls.__name__}", cache_cls)

        return default_mapper



from .candidate import (
    BaseCandidateLayerCache,
    CandidateLayerFileCache,
    CandidateLayerNoCache,
)
from .consts import DEFAULT_CACHE_DIR
from .method import (
    BaseMethodLayerDataCache,
    BaseMethodLayerRankingCache,
    MethodLayerDataFileCache,
    MethodLayerDataNoCache,
    MethodLayerRankingFileCache,
    MethodLayerRankingNoCache,
)
from .styling import BaseStylingLayerCache, StylingLayerFileCache, StylingLayerNoCache
from .xml import BaseXMLLayerCache, XMLLayerFileCache, XMLLayerNoCache

# isort: unique-list
__all__ = [
    "BaseCandidateLayerCache",
    "BaseMethodLayerDataCache",
    "BaseMethodLayerRankingCache",
    "BaseStylingLayerCache",
    "BaseXMLLayerCache",
    "CandidateLayerFileCache",
    "CandidateLayerNoCache",
    "DEFAULT_CACHE_DIR",
    "MethodLayerDataFileCache",
    "MethodLayerDataNoCache",
    "MethodLayerRankingFileCache",
    "MethodLayerRankingNoCache",
    "StylingLayerFileCache",
    "StylingLayerNoCache",
    "XMLLayerFileCache",
    "XMLLayerNoCache",
]



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Set

from py_pdf_term._common.data import Term


@dataclass(frozen=True)
class PageCandidateTermList:
    page_num: int
    candidates: List[Term]

    def to_nostyle_candidates_dict(
        self, to_str: Callable[[Term], str] = str
    ) -> Dict[str, Term]:
        term_dict: Dict[str, Term] = dict()
        for candidate in self.candidates:
            candidate_str = to_str(candidate)
            if candidate_str in term_dict:
                continue
            term_dict[candidate_str] = Term(candidate.tokens, 0.0, "", False)
        return term_dict

    def to_candidates_str_set(self, to_str: Callable[[Term], str] = str) -> Set[str]:
        return {to_str(candidate) for candidate in self.candidates}

    def to_dict(self) -> Dict[str, Any]:
        return {
            "page_num": self.page_num,
            "candidates": list(map(lambda term: term.to_dict(), self.candidates)),
        }

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> "PageCandidateTermList":
        page_num, candidates = obj["page_num"], obj["candidates"]
        return cls(
            page_num,
            list(map(lambda item: Term.from_dict(item), candidates)),
        )


@dataclass(frozen=True)
class PDFCandidateTermList:
    pdf_path: str
    pages: List[PageCandidateTermList]

    def to_nostyle_candidates_dict(
        self, to_str: Callable[[Term], str] = str
    ) -> Dict[str, Term]:
        term_dict: Dict[str, Term] = dict()
        for page in self.pages:
            candidate_items = page.to_nostyle_candidates_dict(to_str).items()
            for candidate_str, candidate in candidate_items:
                if candidate_str in term_dict:
                    continue
                term_dict[candidate_str] = candidate
        return term_dict

    def to_candidates_str_set(self, to_str: Callable[[Term], str] = str) -> Set[str]:
        empty: Set[str] = set()
        return empty.union(
            *map(lambda page: page.to_candidates_str_set(to_str), self.pages)
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "pdf_path": self.pdf_path,
            "pages": list(map(lambda page: page.to_dict(), self.pages)),
        }

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> "PDFCandidateTermList":
        pdf_path, pages = obj["pdf_path"], obj["pages"]
        return cls(
            pdf_path,
            list(map(lambda item: PageCandidateTermList.from_dict(item), pages)),
        )


@dataclass(frozen=True)
class DomainCandidateTermList:
    domain: str
    pdfs: List[PDFCandidateTermList]

    def to_nostyle_candidates_dict(
        self, to_str: Callable[[Term], str] = str
    ) -> Dict[str, Term]:
        term_dict: Dict[str, Term] = dict()
        for pdf in self.pdfs:
            candidate_items = pdf.to_nostyle_candidates_dict(to_str).items()
            for candidate_str, candidate in candidate_items:
                if candidate_str in term_dict:
                    continue
                term_dict[candidate_str] = candidate
        return term_dict

    def to_candidates_str_set(self, to_str: Callable[[Term], str] = str) -> Set[str]:
        empty: Set[str] = set()
        return empty.union(
            *map(lambda pdf: pdf.to_candidates_str_set(to_str), self.pdfs)
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "domain": self.domain,
            "pdfs": list(map(lambda pdf: pdf.to_dict(), self.pdfs)),
        }

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> "DomainCandidateTermList":
        domain, pdfs = obj["domain"], obj["pdfs"]
        return cls(
            domain,
            list(map(lambda item: PDFCandidateTermList.from_dict(item), pdfs)),
        )



from typing import Any, Dict

from .base import BaseMultiDomainRankingMethod
from .collectors import TFIDFRankingDataCollector
from .rankers import TFIDFRanker
from .rankingdata import TFIDFRankingData


class TFIDFMethod(BaseMultiDomainRankingMethod[TFIDFRankingData]):
    def __init__(self, tfmode: str = "log", idfmode: str = "natural") -> None:
        collector = TFIDFRankingDataCollector()
        ranker = TFIDFRanker(tfmode=tfmode, idfmode=idfmode)
        super().__init__(collector, ranker)

    @classmethod
    def collect_data_from_dict(cls, obj: Dict[str, Any]) -> TFIDFRankingData:
        return TFIDFRankingData(**obj)



# pyright:reportUnknownMemberType=false
# pyright:reportUnknownArgumentType=false
# pyright:reportUnknownLambdaType=false

import re
from typing import Any, List

import en_core_web_sm

from py_pdf_term._common.consts import ALPHABET_REGEX, SYMBOL_REGEX

from .data import Token
from .base import BaseLanguageTokenizer


class EnglishTokenizer(BaseLanguageTokenizer):
    def __init__(self) -> None:
        enable_pipes = ["tok2vec", "tagger", "attribute_ruler", "lemmatizer"]
        self._model = en_core_web_sm.load()
        self._model.select_pipes(enable=enable_pipes)

        self._en_regex = re.compile(ALPHABET_REGEX)
        self._symbol_regex = re.compile(rf"({SYMBOL_REGEX})")

    def inscope(self, text: str) -> bool:
        return self._en_regex.search(text) is not None

    def tokenize(self, text: str) -> List[Token]:
        text = self._symbol_regex.sub(r" \1 ", text)
        return list(map(self._create_token, self._model(text)))

    def _create_token(self, token: Any) -> Token:
        if self._symbol_regex.fullmatch(token.text):
            return Token("en", token.text, "SYM", "*", "*", token.text)

        return Token(
            "en", token.text, token.pos_, token.tag_, "*", token.lemma_.lower()
        )



from typing import Any, Dict

from .base import BaseSingleDomainRankingMethod
from .collectors import FLRRankingDataCollector
from .rankers import FLRRanker
from .rankingdata import FLRRankingData


class FLRMethod(BaseSingleDomainRankingMethod[FLRRankingData]):
    def __init__(self) -> None:
        collector = FLRRankingDataCollector()
        ranker = FLRRanker()
        super().__init__(collector, ranker)

    @classmethod
    def collect_data_from_dict(cls, obj: Dict[str, Any]) -> FLRRankingData:
        return FLRRankingData(**obj)



from typing import List, Tuple, Optional

from py_pdf_term._common.data import Term

from .base import BaseSplitter
from ..classifiers import BaseTokenClassifier


class RepeatSplitter(BaseSplitter):
    def __init__(self, classifiers: Optional[List[BaseTokenClassifier]] = None) -> None:
        super().__init__(classifiers=classifiers)

    def split(self, term: Term) -> List[Term]:
        if self._contains_connector_token(term):
            return [term]

        head, backward_splitted_terms = self._backward_split(term)
        forward_splitted_terms, center_term = self._forward_split(head)
        return forward_splitted_terms + [center_term] + backward_splitted_terms

    def _contains_connector_token(self, term: Term) -> bool:
        return any(
            map(
                lambda token: any(
                    map(
                        lambda classifier: classifier.is_connector(token),
                        self._classifiers,
                    )
                ),
                term.tokens,
            )
        )

    def _backward_split(self, term: Term) -> Tuple[Term, List[Term]]:
        splitted_terms: List[Term] = []
        head = term.tokens
        fontsize, ncolor, augmented = term.fontsize, term.ncolor, term.augmented

        while True:
            head_length = len(head)
            j = head_length
            for i in range(head_length - 1, 0, -1):
                if str(head[i - 1]) != str(head[j - 1]):
                    continue
                splitted_term = Term(head[i:j], fontsize, ncolor, augmented)
                splitted_terms.append(splitted_term)
                head = head[:i]
                j = i

            if j == head_length:
                break

        splitted_terms.reverse()
        return Term(head, fontsize, ncolor, augmented), splitted_terms

    def _forward_split(self, term: Term) -> Tuple[List[Term], Term]:
        splitted_terms: List[Term] = []
        tail = term.tokens
        fontsize, ncolor, augmented = term.fontsize, term.ncolor, term.augmented

        while True:
            tail_length = len(tail)
            i = 0
            for j in range(1, tail_length):
                if str(tail[0]) != str(tail[j - i]):
                    continue
                splitted_term = Term(tail[: j - i], fontsize, ncolor, augmented)
                splitted_terms.append(splitted_term)
                tail = tail[j - i :]
                i = j

            if i == 0:
                break

        return splitted_terms, Term(tail, fontsize, ncolor, augmented)



from typing import Type

from py_pdf_term.candidates.filters import (
    BaseCandidateTermFilter,
    BaseCandidateTokenFilter,
    EnglishConcatenationFilter,
    EnglishNumericFilter,
    EnglishProperNounFilter,
    EnglishSymbolLikeFilter,
    EnglishTokenFilter,
    JapaneseConcatenationFilter,
    JapaneseNumericFilter,
    JapaneseProperNounFilter,
    JapaneseSymbolLikeFilter,
    JapaneseTokenFilter,
)

from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class CandidateTokenFilterMapper(BaseMapper[Type[BaseCandidateTokenFilter]]):
    @classmethod
    def default_mapper(cls) -> "CandidateTokenFilterMapper":
        default_mapper = cls()

        token_filter_clses = [JapaneseTokenFilter, EnglishTokenFilter]
        for filter_cls in token_filter_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{filter_cls.__name__}", filter_cls)

        return default_mapper


class CandidateTermFilterMapper(BaseMapper[Type[BaseCandidateTermFilter]]):
    @classmethod
    def default_mapper(cls) -> "CandidateTermFilterMapper":
        default_mapper = cls()

        term_filter_clses = [
            JapaneseConcatenationFilter,
            EnglishConcatenationFilter,
            JapaneseSymbolLikeFilter,
            EnglishSymbolLikeFilter,
            JapaneseProperNounFilter,
            EnglishProperNounFilter,
            JapaneseNumericFilter,
            EnglishNumericFilter,
        ]
        for filter_cls in term_filter_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{filter_cls.__name__}", filter_cls)

        return default_mapper



from dataclasses import dataclass
from typing import Dict

from .base import BaseRankingData


@dataclass(frozen=True)
class TFIDFRankingData(BaseRankingData):
    domain: str
    # unique domain name
    term_freq: Dict[str, int]
    # brute force counting of lemmatized term occurrences in the domain
    # count even if the lemmatized term occurs as a part of a lemmatized phrase
    doc_freq: Dict[str, int]
    # number of documents in the domain that contain the lemmatized term
    # count even if the lemmatized term occurs as a part of a lemmatized phrase
    num_docs: int
    # number of documents in the domain



from typing import Type

from py_pdf_term.tokenizer import (
    BaseLanguageTokenizer,
    EnglishTokenizer,
    JapaneseTokenizer,
)

from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class LanguageTokenizerMapper(BaseMapper[Type[BaseLanguageTokenizer]]):
    @classmethod
    def default_mapper(cls) -> "LanguageTokenizerMapper":
        default_mapper = cls()

        lang_tokenizer_clses = [JapaneseTokenizer, EnglishTokenizer]
        for lang_tokenizer_cls in lang_tokenizer_clses:
            default_mapper.add(
                f"{PACKAGE_NAME}.{lang_tokenizer_cls.__name__}", lang_tokenizer_cls
            )

        return default_mapper



from py_pdf_term.analysis import TermOccurrenceAnalyzer
from py_pdf_term.candidates import DomainCandidateTermList

from ..rankingdata import TFIDFRankingData
from .base import BaseRankingDataCollector


class TFIDFRankingDataCollector(BaseRankingDataCollector[TFIDFRankingData]):
    def __init__(self) -> None:
        super().__init__()
        self._termocc_analyzer = TermOccurrenceAnalyzer()

    def collect(self, domain_candidates: DomainCandidateTermList) -> TFIDFRankingData:
        termocc = self._termocc_analyzer.analyze(domain_candidates)
        num_docs = len(domain_candidates.pdfs)

        return TFIDFRankingData(
            domain_candidates.domain,
            termocc.term_freq,
            termocc.doc_term_freq,
            num_docs,
        )



from .multi import MultiDomainRankingMethodMapper
from .single import SingleDomainRankingMethodMapper

# isort: unique-list
__all__ = ["MultiDomainRankingMethodMapper", "SingleDomainRankingMethodMapper"]



from py_pdf_term._common.data import ScoredTerm, Term
from py_pdf_term.candidates import DomainCandidateTermList

from ..data import MethodTermRanking
from ..rankingdata import FLRHRankingData, FLRRankingData, HITSRankingData
from .base import BaseSingleDomainRanker
from .flr import FLRRanker
from .hits import HITSAuthHubData, HITSRanker


class FLRHRanker(BaseSingleDomainRanker[FLRHRankingData]):
    def __init__(self, threshold: float = 1e-8, max_loop: int = 1000) -> None:
        self._flr_ranker = FLRRanker()
        self._hits_ranker = HITSRanker(threshold=threshold, max_loop=max_loop)

    # pyright:reportPrivateUsage=false
    # FLRHRanker is a friend of FLRRanker and HITSRanker
    def rank_terms(
        self, domain_candidates: DomainCandidateTermList, ranking_data: FLRHRankingData
    ) -> MethodTermRanking:
        flr_ranking_data = FLRRankingData(
            ranking_data.domain,
            ranking_data.term_freq,
            ranking_data.left_freq,
            ranking_data.right_freq,
        )
        hits_ranking_data = HITSRankingData(
            ranking_data.domain,
            ranking_data.term_freq,
            ranking_data.left_freq,
            ranking_data.right_freq,
        )

        auth_hub_data = self._hits_ranker._create_auth_hub_data(hits_ranking_data)
        domain_candidates_dict = domain_candidates.to_nostyle_candidates_dict(
            to_str=lambda candidate: candidate.lemma()
        )
        ranking = list(
            map(
                lambda candidate: self._calculate_score(
                    candidate, flr_ranking_data, hits_ranking_data, auth_hub_data
                ),
                domain_candidates_dict.values(),
            )
        )
        ranking.sort(key=lambda term: -term.score)
        return MethodTermRanking(domain_candidates.domain, ranking)

    def _calculate_score(
        self,
        candidate: Term,
        flr_ranking_data: FLRRankingData,
        hits_ranking_data: HITSRankingData,
        auth_hub_data: HITSAuthHubData,
    ) -> ScoredTerm:
        candidate_lemma = candidate.lemma()
        flr_score = self._flr_ranker._calculate_score(candidate, flr_ranking_data).score
        hits_score = self._hits_ranker._calculate_score(
            candidate, hits_ranking_data, auth_hub_data
        ).score
        return ScoredTerm(candidate_lemma, flr_score + hits_score)



from dataclasses import dataclass
from typing import Dict, Set

from py_pdf_term._common.data import Term
from py_pdf_term.candidates import DomainCandidateTermList

from ..runner import AnalysisRunner


@dataclass(frozen=True)
class DomainTermOccurrence:
    domain: str
    # unique domain name
    term_freq: Dict[str, int]
    # brute force counting of lemmatized term occurrences in the domain
    # count even if the lemmatized term occurs as a part of a lemmatized phrase
    doc_term_freq: Dict[str, int]
    # number of documents in the domain that contain the lemmatized term
    # count even if the lemmatized term occurs as a part of a lemmatized phrase


@dataclass(frozen=True)
class _DomainTermOccurrence:
    domain: str
    # unique domain name
    term_freq: Dict[str, int]
    # brute force counting of lemmatized term occurrences in the domain
    # count even if the term occurs as a part of a lemmatized phrase
    doc_term_set: Dict[str, Set[int]]
    # set of document IDs in the domain that contain the lemmatized term
    # add even if the lemmatized term occurs as a part of a lemmatized phrase


class TermOccurrenceAnalyzer:
    def __init__(self, ignore_augmented: bool = True) -> None:
        self._runner = AnalysisRunner(ignore_augmented=ignore_augmented)

    def analyze(
        self, domain_candidates: DomainCandidateTermList
    ) -> DomainTermOccurrence:
        domain_candidates_set = domain_candidates.to_candidates_str_set(
            lambda candidate: candidate.lemma()
        )

        def update(
            term_occ: _DomainTermOccurrence,
            pdf_id: int,
            page_num: int,
            subcandidate: Term,
        ) -> None:
            subcandidate_lemma = subcandidate.lemma()
            if subcandidate_lemma not in domain_candidates_set:
                return
            term_occ.term_freq[subcandidate_lemma] = (
                term_occ.term_freq.get(subcandidate_lemma, 0) + 1
            )
            doc_term_set = term_occ.doc_term_set.get(subcandidate_lemma, set())
            doc_term_set.add(pdf_id)
            term_occ.doc_term_set[subcandidate_lemma] = doc_term_set

        term_occ = self._runner.run_through_subcandidates(
            domain_candidates,
            _DomainTermOccurrence(domain_candidates.domain, dict(), dict()),
            update,
        )
        term_occ = self._finalize(term_occ)
        return term_occ

    def _finalize(self, term_occ: _DomainTermOccurrence) -> DomainTermOccurrence:
        doc_term_freq = {
            candidate_str: len(doc_term_set)
            for candidate_str, doc_term_set in term_occ.doc_term_set.items()
        }
        return DomainTermOccurrence(term_occ.domain, term_occ.term_freq, doc_term_freq)



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Generic, Iterator, List, Optional

from py_pdf_term.candidates import DomainCandidateTermList

from .collectors import BaseRankingDataCollector
from .data import MethodTermRanking
from .rankers import BaseMultiDomainRanker, BaseSingleDomainRanker
from .rankingdata.base import RankingData


class BaseSingleDomainRankingMethod(Generic[RankingData], metaclass=ABCMeta):
    def __init__(
        self,
        data_collector: BaseRankingDataCollector[RankingData],
        ranker: BaseSingleDomainRanker[RankingData],
    ) -> None:
        self._data_collector = data_collector
        self._ranker = ranker

    def rank_terms(
        self,
        domain_candidates: DomainCandidateTermList,
        ranking_data: Optional[RankingData] = None,
    ) -> MethodTermRanking:
        if ranking_data is None:
            ranking_data = self._data_collector.collect(domain_candidates)
        term_ranking = self._ranker.rank_terms(domain_candidates, ranking_data)
        return term_ranking

    def collect_data(self, domain_candidates: DomainCandidateTermList) -> RankingData:
        ranking_data = self._data_collector.collect(domain_candidates)
        return ranking_data

    @classmethod
    @abstractmethod
    def collect_data_from_dict(cls, obj: Dict[str, Any]) -> RankingData:
        raise NotImplementedError(f"{cls.__name__}.collect_data_from_dict()")


class BaseMultiDomainRankingMethod(Generic[RankingData], metaclass=ABCMeta):
    def __init__(
        self,
        data_collector: BaseRankingDataCollector[RankingData],
        ranker: BaseMultiDomainRanker[RankingData],
    ) -> None:
        self._data_collector = data_collector
        self._ranker = ranker

    def rank_terms(
        self,
        domain_candidates_list: List[DomainCandidateTermList],
        ranking_data_list: Optional[List[RankingData]] = None,
    ) -> Iterator[MethodTermRanking]:
        if ranking_data_list is None:
            ranking_data_list = list(
                map(self._data_collector.collect, domain_candidates_list)
            )

        for domain_candidates in domain_candidates_list:
            term_ranking = self._ranker.rank_terms(domain_candidates, ranking_data_list)
            yield term_ranking

    def rank_domain_terms(
        self,
        domain: str,
        domain_candidates_list: List[DomainCandidateTermList],
        ranking_data_list: Optional[List[RankingData]] = None,
    ) -> MethodTermRanking:
        domain_candidates = next(
            filter(lambda item: item.domain == domain, domain_candidates_list)
        )

        if ranking_data_list is None:
            ranking_data_list = list(
                map(self._data_collector.collect, domain_candidates_list)
            )

        term_ranking = self._ranker.rank_terms(domain_candidates, ranking_data_list)
        return term_ranking

    def collect_data(self, domain_candidates: DomainCandidateTermList) -> RankingData:
        ranking_data = self._data_collector.collect(domain_candidates)
        return ranking_data

    @classmethod
    @abstractmethod
    def collect_data_from_dict(cls, obj: Dict[str, Any]) -> RankingData:
        raise NotImplementedError(f"{cls.__name__}.collect_data_from_dict()")



from dataclasses import dataclass, field
from typing import Any, Dict

from .base import BaseLayerConfig


@dataclass(frozen=True)
class MethodLayerConfig(BaseLayerConfig):
    method_type: str = "single"
    method: str = "py_pdf_term.FLRHMethod"
    hyper_params: Dict[str, Any] = field(default_factory=dict)
    ranking_cache: str = "py_pdf_term.MethodLayerRankingFileCache"
    data_cache: str = "py_pdf_term.MethodLayerRankingFileCache"



import re

from py_pdf_term._common.consts import ALPHABET_REGEX, HIRAGANA_REGEX, KATAKANA_REGEX
from py_pdf_term._common.data import Term
from ....classifiers import JapaneseTokenClassifier

from ..base import BaseJapaneseCandidateTermFilter

PHONETIC_REGEX = rf"{HIRAGANA_REGEX}|{KATAKANA_REGEX}|{ALPHABET_REGEX}"


class JapaneseConcatenationFilter(BaseJapaneseCandidateTermFilter):
    def __init__(self) -> None:
        self._classifier = JapaneseTokenClassifier()

    def is_candidate(self, scoped_term: Term) -> bool:
        return (
            self._is_norn_phrase(scoped_term)
            and not self._has_invalid_connector_symbol(scoped_term)
            and not self._has_invalid_connector_term(scoped_term)
            and not self._has_invalid_prefix(scoped_term)
            and not self._has_invalid_postfix(scoped_term)
            and not self._has_invalid_adjective(scoped_term)
            and not self._has_invalid_verb(scoped_term)
        )

    def _is_norn_phrase(self, scoped_term: Term) -> bool:
        num_tokens = len(scoped_term.tokens)

        def norn_or_postfix_appears_at(i: int) -> bool:
            return scoped_term.tokens[i].pos in {"名詞", "記号", "接尾辞"}

        induces_should_be_norn = [
            i - 1
            for i in range(1, num_tokens)
            if self._classifier.is_connector_term(scoped_term.tokens[i])
        ] + [num_tokens - 1]

        return all(map(norn_or_postfix_appears_at, induces_should_be_norn))

    def _has_invalid_connector_symbol(self, scoped_term: Term) -> bool:
        num_tokens = len(scoped_term.tokens)

        def invalid_connector_symbol_appears_at(i: int) -> bool:
            if not self._classifier.is_connector_symbol(scoped_term.tokens[i]):
                return False
            return (
                i == 0
                or i == num_tokens - 1
                or self._classifier.is_connector_symbol(scoped_term.tokens[i - 1])
                or self._classifier.is_connector_symbol(scoped_term.tokens[i + 1])
            )

        return any(map(invalid_connector_symbol_appears_at, range(num_tokens)))

    def _has_invalid_connector_term(self, scoped_term: Term) -> bool:
        num_tokens = len(scoped_term.tokens)
        phonetic_regex = re.compile(PHONETIC_REGEX)

        def invalid_connector_symbol_appears_at(i: int) -> bool:
            if not self._classifier.is_connector_term(scoped_term.tokens[i]):
                return False
            return (
                i == 0
                or i == num_tokens - 1
                or self._classifier.is_connector_term(scoped_term.tokens[i - 1])
                or self._classifier.is_connector_term(scoped_term.tokens[i + 1])
                or self._classifier.is_symbol(scoped_term.tokens[i - 1])
                or self._classifier.is_symbol(scoped_term.tokens[i + 1])
                or phonetic_regex.fullmatch(str(scoped_term.tokens[i - 1])) is not None
                or phonetic_regex.fullmatch(str(scoped_term.tokens[i + 1])) is not None
            )

        return any(map(invalid_connector_symbol_appears_at, range(num_tokens)))

    def _has_invalid_prefix(self, scoped_term: Term) -> bool:
        num_tokens = len(scoped_term.tokens)

        def invalid_prefix_appears_at(i: int) -> bool:
            token = scoped_term.tokens[i]
            if token.pos != "接頭辞":
                return False
            return (
                i == num_tokens - 1
                or scoped_term.tokens[i + 1].pos not in {"名詞", "記号", "形状詞"}
                # No line break
            )

        return any(map(invalid_prefix_appears_at, range(num_tokens)))

    def _has_invalid_postfix(self, scoped_term: Term) -> bool:
        num_tokens = len(scoped_term.tokens)

        def invalid_postfix_appears_at(i: int) -> bool:
            token = scoped_term.tokens[i]
            if token.pos != "接尾辞":
                return False
            return (
                i == 0
                or scoped_term.tokens[i - 1].pos not in {"名詞", "記号", "形状詞", "動詞", "形容詞"}
                # No line break
            )

        return any(map(invalid_postfix_appears_at, range(num_tokens)))

    def _has_invalid_adjective(self, scoped_term: Term) -> bool:
        num_tokens = len(scoped_term.tokens)

        def invalid_adjective_appears_at(i: int) -> bool:
            token = scoped_term.tokens[i]
            if token.pos not in {"形状詞", "形容詞"}:
                return False
            return (
                token.category == ""
                or i == num_tokens - 1
                or scoped_term.tokens[i + 1].pos
                not in {"名詞", "記号", "接尾辞", "形状詞", "形容詞"}
            )

        return any(map(invalid_adjective_appears_at, range(num_tokens)))

    def _has_invalid_verb(self, scoped_term: Term) -> bool:
        num_tokens = len(scoped_term.tokens)

        def invalid__verb_appears_at(i: int) -> bool:
            token = scoped_term.tokens[i]
            if token.pos != "動詞":
                return False
            return (
                i == num_tokens - 1
                or scoped_term.tokens[i + 1].pos not in {"接尾辞", "動詞"}
                # No line break
            )

        return any(map(invalid__verb_appears_at, range(num_tokens)))



from py_pdf_term.tokenizer import Token
from .base import BaseTokenClassifier


class JapaneseTokenClassifier(BaseTokenClassifier):
    def inscope(self, token: Token) -> bool:
        return token.lang == "ja"

    def is_symbol(self, token: Token) -> bool:
        return token.pos in {"補助記号"}

    def is_connector_symbol(self, token: Token) -> bool:
        return token.surface_form in {"・", "-"} and token.pos == "補助記号"

    def is_connector_term(self, token: Token) -> bool:
        return token.surface_form == "の" and token.pos == "助詞"



from .base import BaseAugmenter
from .combiner import AugmenterCombiner
from .separation import EnglishConnectorTermAugmenter, JapaneseConnectorTermAugmenter

# isort: unique-list
__all__ = [
    "AugmenterCombiner",
    "BaseAugmenter",
    "EnglishConnectorTermAugmenter",
    "JapaneseConnectorTermAugmenter",
]



from typing import Union

from py_pdf_term.stylings import PDFStylingScoreList

from ...configs import StylingLayerConfig
from .base import BaseStylingLayerCache


class StylingLayerNoCache(BaseStylingLayerCache):
    def __init__(self, cache_dir: str) -> None:
        pass

    def load(
        self, pdf_path: str, config: StylingLayerConfig
    ) -> Union[PDFStylingScoreList, None]:
        pass

    def store(
        self, styling_scores: PDFStylingScoreList, config: StylingLayerConfig
    ) -> None:
        pass

    def remove(self, pdf_path: str, config: StylingLayerConfig) -> None:
        pass



from typing import Any, Dict

from .base import BaseSingleDomainRankingMethod
from .collectors import FLRHRankingDataCollector
from .rankers import FLRHRanker
from .rankingdata import FLRHRankingData


class FLRHMethod(BaseSingleDomainRankingMethod[FLRHRankingData]):
    def __init__(self, threshold: float = 1e-8, max_loop: int = 1000) -> None:
        collector = FLRHRankingDataCollector()
        ranker = FLRHRanker(threshold=threshold, max_loop=max_loop)
        super().__init__(collector, ranker)

    @classmethod
    def collect_data_from_dict(cls, obj: Dict[str, Any]) -> FLRHRankingData:
        return FLRHRankingData(**obj)



from .base import BaseLayerConfig
from .candidate import CandidateLayerConfig
from .method import MethodLayerConfig
from .styling import StylingLayerConfig
from .techterm import TechnicalTermLayerConfig
from .xml import XMLLayerConfig

# isort: unique-list
__all__ = [
    "BaseLayerConfig",
    "CandidateLayerConfig",
    "MethodLayerConfig",
    "StylingLayerConfig",
    "TechnicalTermLayerConfig",
    "XMLLayerConfig",
]



from .base import BaseSplitter
from .combiner import SplitterCombiner
from .repeat import RepeatSplitter
from .symname import SymbolNameSplitter

# isort: unique-list
__all__ = ["BaseSplitter", "RepeatSplitter", "SplitterCombiner", "SymbolNameSplitter"]



from math import log10
from typing import List

from py_pdf_term._common.data import ScoredTerm, Term
from py_pdf_term._common.extended_math import extended_log10
from py_pdf_term.candidates import DomainCandidateTermList

from ..data import MethodTermRanking
from ..rankingdata import TFIDFRankingData
from .base import BaseMultiDomainRanker


class TFIDFRanker(BaseMultiDomainRanker[TFIDFRankingData]):
    def __init__(self, tfmode: str = "log", idfmode: str = "natural") -> None:
        self._tfmode = tfmode
        self._idfmode = idfmode

    def rank_terms(
        self,
        domain_candidates: DomainCandidateTermList,
        ranking_data_list: List[TFIDFRankingData],
    ) -> MethodTermRanking:
        domain_candidates_dict = domain_candidates.to_nostyle_candidates_dict()
        ranking_data = next(
            filter(
                lambda item: item.domain == domain_candidates.domain,
                ranking_data_list,
            )
        )
        ranking = list(
            map(
                lambda candidate: self._calculate_score(
                    candidate, ranking_data, ranking_data_list
                ),
                domain_candidates_dict.values(),
            )
        )
        ranking.sort(key=lambda term: -term.score)
        return MethodTermRanking(domain_candidates.domain, ranking)

    def _calculate_score(
        self,
        candidate: Term,
        ranking_data: TFIDFRankingData,
        ranking_data_list: List[TFIDFRankingData],
    ) -> ScoredTerm:
        candidate_lemma = candidate.lemma()

        tf = self._calculate_tf(candidate_lemma, ranking_data, ranking_data_list)
        idf = self._calculate_idf(candidate_lemma, ranking_data, ranking_data_list)
        score = extended_log10(tf * idf)
        return ScoredTerm(candidate_lemma, score)

    def _calculate_tf(
        self,
        candidate: str,
        ranking_data: TFIDFRankingData,
        ranking_data_list: List[TFIDFRankingData],
    ) -> float:
        tf = ranking_data.term_freq.get(candidate, 0)

        if self._idfmode == "natural":
            return tf

        elif self._tfmode == "log":
            return 1.0 * log10(tf) if tf > 0 else 0.0

        elif self._tfmode == "augmented":
            max_tf = max(
                map(lambda data: data.term_freq.get(candidate, 0), ranking_data_list)
            )
            return 0.5 + 0.5 * tf / max_tf if max_tf > 0 else 0.0

        elif self._tfmode == "logave":
            ave_tf = sum(
                map(lambda data: data.term_freq.get(candidate, 0), ranking_data_list)
            ) / len(ranking_data_list)
            return (
                (1.0 + log10(tf)) / (1.0 + log10(ave_tf))
                if tf > 0 and ave_tf > 0.0
                else 0.0
            )

        elif self._tfmode == "binary":
            return 1.0 if tf > 0 else 0.0

        raise ValueError(f"unknown tfmode {self._tfmode}")

    def _calculate_idf(
        self,
        candidate: str,
        ranking_data: TFIDFRankingData,
        ranking_data_list: List[TFIDFRankingData],
    ) -> float:
        num_docs = sum(map(lambda data: data.num_docs, ranking_data_list))
        df = sum(map(lambda data: data.doc_freq.get(candidate, 0), ranking_data_list))

        if self._idfmode == "natural":
            return log10(num_docs / df) if df > 0 else 0.0

        if self._idfmode == "smooth":
            return log10(num_docs / (df + 1)) + 1.0

        elif self._idfmode == "prob":
            return max(log10((num_docs - df) / df), 0.0) if df > 0 else 0.0

        elif self._idfmode == "unary":
            return 1.0

        raise ValueError(f"unknown idfmode {self._idfmode}")



from .candidate import CandidateLayer
from .method import MethodLayer
from .styling import StylingLayer
from .techterm import TechnicalTermLayer
from .xml import XMLLayer

# isort: unique-list
__all__ = [
    "CandidateLayer",
    "MethodLayer",
    "StylingLayer",
    "TechnicalTermLayer",
    "XMLLayer",
]



from typing import Type

from py_pdf_term.candidates.classifiers import (
    BaseTokenClassifier,
    EnglishTokenClassifier,
    JapaneseTokenClassifier,
)

from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class TokenClassifilerMapper(BaseMapper[Type[BaseTokenClassifier]]):
    @classmethod
    def default_mapper(cls) -> "TokenClassifilerMapper":
        default_mapper = cls()

        classifier_clses = [
            JapaneseTokenClassifier,
            EnglishTokenClassifier,
        ]
        for classifier_cls in classifier_clses:
            default_mapper.add(
                f"{PACKAGE_NAME}.{classifier_cls.__name__}", classifier_cls
            )

        return default_mapper



from typing import List, Optional

from py_pdf_term._common.data import Term

from .base import BaseAugmenter
from .separation import EnglishConnectorTermAugmenter, JapaneseConnectorTermAugmenter


class AugmenterCombiner:
    def __init__(self, augmenters: Optional[List[BaseAugmenter]] = None) -> None:
        if augmenters is None:
            augmenters = [
                JapaneseConnectorTermAugmenter(),
                EnglishConnectorTermAugmenter(),
            ]

        self._augmenters = augmenters

    def augment(self, term: Term) -> List[Term]:
        augmented_terms = [term]
        for augmenter in self._augmenters:
            start: List[Term] = []
            augmented_terms += sum(map(augmenter.augment, augmented_terms), start)

        return augmented_terms[1:]



import re
from typing import List

from py_pdf_term._common.consts import ENGLISH_REGEX, NUMBER_REGEX
from py_pdf_term.tokenizer import Token

from .base import BaseCandidateTokenFilter


class EnglishTokenFilter(BaseCandidateTokenFilter):
    def __init__(self) -> None:
        self._regex = re.compile(rf"({ENGLISH_REGEX}|{NUMBER_REGEX})+")

    def inscope(self, token: Token) -> bool:
        token_str = str(token)
        return token.lang == "en" and (
            self._regex.fullmatch(token_str) is not None or token_str == "-"
        )

    def is_partof_candidate(self, tokens: List[Token], idx: int) -> bool:
        scoped_token = tokens[idx]
        num_tokens = len(tokens)

        if scoped_token.pos in {"NOUN", "PROPN", "NUM"}:
            return True
        elif scoped_token.pos == "ADJ":
            return (
                idx < num_tokens - 1
                and tokens[idx + 1].pos in {"NOUN", "PROPN", "ADJ", "VERB", "SYM"}
                # No line break
            )
        elif scoped_token.pos == "VERB":
            return scoped_token.category == "VBG" or (
                scoped_token.category == "VBN"
                and idx < num_tokens - 1
                and tokens[idx + 1].pos in {"NOUN", "PROPN", "ADJ", "VERB", "SYM"}
            )
        elif scoped_token.pos == "ADP":
            return scoped_token.category == "IN"
        elif scoped_token.pos == "SYM":
            return (
                scoped_token.surface_form == "-"
                and 0 < idx < num_tokens - 1
                and self._regex.match(str(tokens[idx - 1])) is not None
                and self._regex.match(str(tokens[idx + 1])) is not None
            )

        return False



from typing import Any, List, Optional

from py_pdf_term.candidates import DomainCandidateTermList
from py_pdf_term.methods import (
    BaseMultiDomainRankingMethod,
    BaseSingleDomainRankingMethod,
    MethodTermRanking,
)

from ..caches import DEFAULT_CACHE_DIR
from ..configs import MethodLayerConfig
from ..data import DomainPDFList
from ..mappers import (
    MethodLayerDataCacheMapper,
    MethodLayerRankingCacheMapper,
    MultiDomainRankingMethodMapper,
    SingleDomainRankingMethodMapper,
)
from .candidate import CandidateLayer


class MethodLayer:
    def __init__(
        self,
        candidate_layer: CandidateLayer,
        config: Optional[MethodLayerConfig] = None,
        single_method_mapper: Optional[SingleDomainRankingMethodMapper] = None,
        multi_method_mapper: Optional[MultiDomainRankingMethodMapper] = None,
        ranking_cache_mapper: Optional[MethodLayerRankingCacheMapper] = None,
        data_cache_mapper: Optional[MethodLayerDataCacheMapper] = None,
        cache_dir: str = DEFAULT_CACHE_DIR,
    ) -> None:
        if config is None:
            config = MethodLayerConfig()
        if single_method_mapper is None:
            single_method_mapper = SingleDomainRankingMethodMapper.default_mapper()
        if multi_method_mapper is None:
            multi_method_mapper = MultiDomainRankingMethodMapper.default_mapper()
        if ranking_cache_mapper is None:
            ranking_cache_mapper = MethodLayerRankingCacheMapper.default_mapper()
        if data_cache_mapper is None:
            data_cache_mapper = MethodLayerDataCacheMapper.default_mapper()

        if config.method_type == "single":
            method_cls = single_method_mapper.find(config.method)
        elif config.method_type == "multi":
            method_cls = multi_method_mapper.find(config.method)
        else:
            raise ValueError(f"unknown method type '{config.method_type}'")

        ranking_cache_cls = ranking_cache_mapper.find(config.ranking_cache)
        data_cache_cls = data_cache_mapper.find(config.data_cache)

        self._method = method_cls(**config.hyper_params)
        self._ranking_cache = ranking_cache_cls(cache_dir=cache_dir)
        self._data_cache = data_cache_cls(cache_dir=cache_dir)
        self._config = config

        self._candidate_layer = candidate_layer

    # pyright:reportUnnecessaryIsInstance=false
    def create_term_ranking(
        self,
        domain: str,
        single_domain_pdfs: Optional[DomainPDFList] = None,
        multi_domain_pdfs: Optional[List[DomainPDFList]] = None,
    ) -> MethodTermRanking:

        if isinstance(self._method, BaseSingleDomainRankingMethod):
            if single_domain_pdfs is None:
                raise ValueError(
                    "'single_domain_pdfs' is required"
                    "when using single-domain ranking method"
                )
            term_ranking = self._run_single_domain_method(domain, single_domain_pdfs)
            return term_ranking

        elif isinstance(self._method, BaseMultiDomainRankingMethod):
            if multi_domain_pdfs is None:
                raise ValueError(
                    "'multi_domain_pdfs' is required"
                    " when using multi-domain ranking method"
                )
            term_ranking = self._run_multi_domain_method(domain, multi_domain_pdfs)
            return term_ranking

        else:
            raise RuntimeError("unreachable statement")

    def remove_cache(self, pdf_paths: List[str]) -> None:
        self._ranking_cache.remove(pdf_paths, self._config)
        self._data_cache.remove(pdf_paths, self._config)

    def _run_single_domain_method(
        self,
        domain: str,
        single_domain_pdfs: DomainPDFList,
    ) -> MethodTermRanking:
        if not isinstance(self._method, BaseSingleDomainRankingMethod):
            raise RuntimeError("unreachable statement")

        if domain != single_domain_pdfs.domain:
            raise ValueError(
                f"domain of 'single_domain_pdfs is expected to be '{domain}'"
                f" but got '{single_domain_pdfs.domain}'"
            )

        term_ranking = self._ranking_cache.load(
            single_domain_pdfs.pdf_paths, self._config
        )

        if term_ranking is None:
            candidates = self._candidate_layer.create_domain_candiates(
                single_domain_pdfs
            )
            ranking_data = self._create_ranking_data(single_domain_pdfs, candidates)
            term_ranking = self._method.rank_terms(candidates, ranking_data)

        self._ranking_cache.store(
            single_domain_pdfs.pdf_paths, term_ranking, self._config
        )

        return term_ranking

    def _run_multi_domain_method(
        self,
        domain: str,
        multi_domain_pdfs: List[DomainPDFList],
    ) -> MethodTermRanking:
        if not isinstance(self._method, BaseMultiDomainRankingMethod):
            raise RuntimeError("unreachable statement")

        target_domain_pdfs = next(
            filter(lambda item: item.domain == domain, multi_domain_pdfs), None
        )
        if target_domain_pdfs is None:
            raise ValueError(f"'multi_domain_pdfs' does not contain domain '{domain}'")

        term_ranking = self._ranking_cache.load(
            target_domain_pdfs.pdf_paths, self._config
        )

        if term_ranking is None:
            domain_candidates_list: List[DomainCandidateTermList] = []
            ranking_data_list: List[Any] = []
            for domain_pdfs in multi_domain_pdfs:
                candidates = self._candidate_layer.create_domain_candiates(domain_pdfs)
                ranking_data = self._create_ranking_data(domain_pdfs, candidates)
                domain_candidates_list.append(candidates)
                ranking_data_list.append(ranking_data)

            term_ranking = self._method.rank_domain_terms(
                domain, domain_candidates_list, ranking_data_list
            )

        self._ranking_cache.store(
            target_domain_pdfs.pdf_paths, term_ranking, self._config
        )

        return term_ranking

    def _create_ranking_data(
        self, domain_pdfs: DomainPDFList, domain_candidates: DomainCandidateTermList
    ) -> Any:
        ranking_data = self._data_cache.load(
            domain_pdfs.pdf_paths,
            self._config,
            self._method.collect_data_from_dict,
        )

        if ranking_data is None:
            ranking_data = self._method.collect_data(domain_candidates)

        self._data_cache.store(domain_pdfs.pdf_paths, ranking_data, self._config)

        return ranking_data



from dataclasses import asdict, dataclass
from typing import Any, ClassVar, Dict


@dataclass
class Token:
    NUM_ATTR: ClassVar[int] = 6

    lang: str
    surface_form: str
    pos: str
    category: str
    subcategory: str
    lemma: str
    is_meaningless: bool = False

    def __str__(self) -> str:
        return self.surface_form

    def to_dict(self) -> Dict[str, str]:
        return asdict(self)

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> "Token":
        return cls(**obj)



from typing import Type

from ...caches import (
    BaseCandidateLayerCache,
    CandidateLayerFileCache,
    CandidateLayerNoCache,
)
from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class CandidateLayerCacheMapper(BaseMapper[Type[BaseCandidateLayerCache]]):
    @classmethod
    def default_mapper(cls) -> "CandidateLayerCacheMapper":
        default_mapper = cls()

        cache_clses = [CandidateLayerNoCache, CandidateLayerFileCache]
        for cache_cls in cache_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{cache_cls.__name__}", cache_cls)

        return default_mapper



from abc import ABCMeta
from typing import Callable, List

from py_pdf_term._common.data import Term
from py_pdf_term.tokenizer import Token
from ..classifiers import EnglishTokenClassifier, JapaneseTokenClassifier

from .base import BaseAugmenter


class BaseSeparationAugmenter(BaseAugmenter, metaclass=ABCMeta):
    def __init__(self, is_separator: Callable[[Token], bool]) -> None:
        self._is_separator = is_separator

    def augment(self, term: Term) -> List[Term]:
        num_tokens = len(term.tokens)
        separation_positions = (
            [-1]
            + [i for i in range(num_tokens) if self._is_separator(term.tokens[i])]
            + [num_tokens]
        )
        num_positions = len(separation_positions)

        augmented_terms: List[Term] = []
        for length in range(1, num_positions - 1):
            for idx in range(num_positions - length):
                i = separation_positions[idx]
                j = separation_positions[idx + length]
                tokens = term.tokens[i + 1 : j]
                augmented_term = Term(tokens, term.fontsize, term.ncolor, True)
                augmented_terms.append(augmented_term)

        return augmented_terms


class JapaneseConnectorTermAugmenter(BaseSeparationAugmenter):
    def __init__(self) -> None:
        classifier = JapaneseTokenClassifier()
        super().__init__(classifier.is_connector_term)

    def augment(self, term: Term) -> List[Term]:
        if term.lang != "ja":
            return []

        return super().augment(term)


class EnglishConnectorTermAugmenter(BaseSeparationAugmenter):
    def __init__(self) -> None:
        classifier = EnglishTokenClassifier()
        super().__init__(classifier.is_connector_term)

    def augment(self, term: Term) -> List[Term]:
        if term.lang != "en":
            return []

        return super().augment(term)



from .endtoend._endtoend.configs import (
    CandidateLayerConfig,
    MethodLayerConfig,
    StylingLayerConfig,
    TechnicalTermLayerConfig,
    XMLLayerConfig,
)

# isort: unique-list
__all__ = [
    "CandidateLayerConfig",
    "MethodLayerConfig",
    "StylingLayerConfig",
    "TechnicalTermLayerConfig",
    "XMLLayerConfig",
]



from typing import BinaryIO, Callable

from ..base import BaseMapper


class BinaryOpenerMapper(BaseMapper[Callable[[str, str], BinaryIO]]):
    @classmethod
    def default_mapper(cls) -> "BinaryOpenerMapper":
        default_mapper = cls()
        default_mapper.add("python.open", open)  # type:ignore
        return default_mapper



from abc import ABCMeta, abstractmethod
from typing import List

from py_pdf_term._common.data import Term


class BaseAugmenter(metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def augment(self, term: Term) -> List[Term]:
        raise NotImplementedError(f"{self.__class__.__name__}.augment()")



from typing import Type

from py_pdf_term.candidates.splitters import (
    BaseSplitter,
    RepeatSplitter,
    SymbolNameSplitter,
)

from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class SplitterMapper(BaseMapper[Type[BaseSplitter]]):
    @classmethod
    def default_mapper(cls) -> "SplitterMapper":
        default_mapper = cls()

        splitter_clses = [SymbolNameSplitter, RepeatSplitter]
        for splitter_cls in splitter_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{splitter_cls.__name__}", splitter_cls)

        return default_mapper



from typing import List, Optional

from py_pdf_term._common.data import Term
from .base import BaseTokenClassifier
from .japanese import JapaneseTokenClassifier
from .english import EnglishTokenClassifier


class MeaninglessMarker:
    def __init__(self, classifiers: Optional[List[BaseTokenClassifier]] = None) -> None:
        if classifiers is None:
            classifiers = [
                JapaneseTokenClassifier(),
                EnglishTokenClassifier(),
            ]

        self._classifiers = classifiers

    def mark(self, term: Term) -> Term:
        for token in term.tokens:
            token.is_meaningless = any(
                map(
                    lambda classifier: classifier.inscope(token)
                    and classifier.is_meaningless(token),
                    self._classifiers,
                )
            )
        return term



from abc import ABCMeta, abstractmethod
from typing import Union

from py_pdf_term.stylings import PDFStylingScoreList

from ...configs import StylingLayerConfig


class BaseStylingLayerCache(metaclass=ABCMeta):
    def __init__(self, cache_dir: str) -> None:
        pass

    @abstractmethod
    def load(
        self, pdf_path: str, config: StylingLayerConfig
    ) -> Union[PDFStylingScoreList, None]:
        raise NotImplementedError(f"{self.__class__.__name__}.load()")

    @abstractmethod
    def store(
        self, styling_scores: PDFStylingScoreList, config: StylingLayerConfig
    ) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.store()")

    @abstractmethod
    def remove(self, pdf_path: str, config: StylingLayerConfig) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.remove()")



# pyright:reportGeneralTypeIssues=false
# pyright:reportUnknownMemberType=false
# pyright:reportUnknownParameterType=false
# pyright:reportUnknownArgumentType=false
# pyright:reportMissingParameterType=false
# pyright:reportMissingTypeArgument=false
# pyright:reportIncompatibleMethodOverride=false

from dataclasses import dataclass
from typing import Any, BinaryIO, Optional, Union

from pdfminer.converter import PDFConverter
from pdfminer.layout import (
    LAParams,
    LTAnno,
    LTChar,
    LTPage,
    LTText,
    LTTextBox,
    LTTextLine,
)
from pdfminer.pdfinterp import PDFResourceManager
from pdfminer.utils import bbox2str, enc

from .utils import clean_content_text


@dataclass
class TextboxState:
    within_section: bool
    size: float
    ncolor: str
    bbox: str
    text: str


class TextfulXMLConverter(PDFConverter):
    def __init__(
        self,
        rsrcmgr: PDFResourceManager,
        outfp: BinaryIO,
        codec: str = "utf-8",
        pageno: int = 1,
        laparams: Optional[LAParams] = None,
        nfc_norm: bool = True,
        include_pattern: Optional[str] = None,
        exclude_pattern: Optional[str] = None,
    ) -> None:
        PDFConverter.__init__(self, rsrcmgr, outfp, codec, pageno, laparams)

        def _clean_content_text(text: str) -> str:
            return clean_content_text(text, nfc_norm, include_pattern, exclude_pattern)

        self._clean_content_text = _clean_content_text

    def write_header(self) -> None:
        if self.codec:
            self._write('<?xml version="1.0" encoding="%s" ?>\n' % self.codec)
        else:
            self._write('<?xml version="1.0" ?>\n')

        self._write("<pages>\n")

    def receive_layout(self, ltpage: LTPage) -> None:
        self._render(ltpage)

    def write_footer(self) -> None:
        self._write("</pages>\n")

    # override to ignore LTFigure
    def begin_figure(self, name, bbox, matrix) -> None:
        pass

    # override to ignore LTFigure
    def end_figure(self, name) -> None:
        pass

    # override to ignore LTImage
    def render_image(self, name, stream) -> None:
        pass

    # override to ignore LTLine, LTRect and LTCurve
    def paint_path(self, graphicstate, stroke, fill, evenodd, path) -> None:
        pass

    def _render(self, item: Any) -> None:
        if isinstance(item, LTPage):
            self._render_page(item)
        elif isinstance(item, LTTextBox):
            self._render_textbox(item)
        elif isinstance(item, LTText):
            self._render_text(item)

    def _render_page(self, ltpage: LTPage) -> None:
        self._write('<page id="%s">\n' % ltpage.pageid)
        for child in ltpage:
            self._render(child)
        self._write("</page>\n")

    def _render_textbox(self, lttextbox: LTTextBox) -> None:
        state = TextboxState(False, 0.0, "", "", "")

        def render_textbox_child(child: Any) -> None:
            if isinstance(child, LTTextLine):
                for grandchild in child:
                    render_textbox_child(grandchild)
            elif isinstance(child, LTChar):
                if not state.within_section:
                    enter_text_section(child)
                    state.text += child.get_text()
                elif text_section_continues(child):
                    state.text += child.get_text()
                else:
                    exit_text_section()
                    enter_text_section(child)
                    state.text += child.get_text()
            elif isinstance(child, LTAnno):
                if not state.within_section:
                    pass
                elif text_section_continues(child):
                    state.text += child.get_text()
                else:
                    exit_text_section()

        def enter_text_section(item: Union[LTChar, LTAnno]) -> None:
            state.within_section = True
            state.size = item.size
            state.ncolor = item.graphicstate.ncolor
            state.bbox = bbox2str(item.bbox)
            state.text = ""

        def text_section_continues(item: Union[LTChar, LTAnno]) -> bool:
            if isinstance(item, LTAnno):
                return True
            return (
                state.ncolor == item.graphicstate.ncolor
                and abs(state.size - item.size) < 0.1
            )

        def exit_text_section() -> None:
            if not state.within_section:
                return

            text = self._clean_content_text(state.text)
            if text:
                self._write(
                    '<text size="%.3f" ncolor="%s" bbox="%s">'
                    % (state.size, state.ncolor, state.bbox)
                )
                self._write(enc(text))
                self._write("</text>\n")

            state.within_section = False
            state.size = 0.0
            state.ncolor = ""
            state.bbox = ""
            state.text = ""

        for child in lttextbox:
            render_textbox_child(child)

        exit_text_section()

    def _render_text(self, lttext: LTText) -> None:
        text = self._clean_content_text(lttext.get_text())
        if text:
            self._write("<text>")
            self._write(enc(text))
            self._write("</text>\n")

    def _write(self, text: str) -> None:
        if self.codec:
            text = text.encode(self.codec)
        self.outfp.write(text)



from typing import List, Optional

from py_pdf_term.techterms import PDFTechnicalTermList, TechnicalTermExtractor

from ..configs import TechnicalTermLayerConfig
from ..data import DomainPDFList
from .candidate import CandidateLayer
from .method import MethodLayer
from .styling import StylingLayer


class TechnicalTermLayer:
    def __init__(
        self,
        candidate_layer: CandidateLayer,
        method_layer: MethodLayer,
        styling_layer: StylingLayer,
        config: Optional[TechnicalTermLayerConfig] = None,
    ) -> None:
        if config is None:
            config = TechnicalTermLayerConfig()

        self._techterm = TechnicalTermExtractor(
            max_num_terms=config.max_num_terms,
            acceptance_rate=config.acceptance_rate,
        )
        self._config = config

        self._candidate_layer = candidate_layer
        self._method_layer = method_layer
        self._styling_layer = styling_layer

    def create_pdf_techterms(
        self,
        domain: str,
        pdf_path: str,
        single_domain_pdfs: Optional[DomainPDFList] = None,
        multi_domain_pdfs: Optional[List[DomainPDFList]] = None,
    ) -> PDFTechnicalTermList:
        pdf_candidates = self._candidate_layer.create_pdf_candidates(pdf_path)
        term_ranking = self._method_layer.create_term_ranking(
            domain, single_domain_pdfs, multi_domain_pdfs
        )
        pdf_styling_scores = self._styling_layer.create_pdf_styling_scores(pdf_path)
        techterms = self._techterm.extract_from_pdf(
            pdf_candidates, term_ranking, pdf_styling_scores
        )

        return techterms



from dataclasses import asdict, dataclass
from typing import Any, Dict, Type, TypeVar

CACHE_CONFIGS = ["cache", "data_cache", "ranking_cache"]


@dataclass(frozen=True)
class BaseLayerConfig:
    pass

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def to_dict_without_cache(self) -> Dict[str, Any]:
        config_dict = asdict(self)
        for cache_config in CACHE_CONFIGS:
            config_dict.pop(cache_config, None)
        return config_dict

    @classmethod
    def from_dict(cls: Type["LayerConfig"], obj: Dict[str, Any]) -> "LayerConfig":
        return cls(**obj)


LayerConfig = TypeVar("LayerConfig", bound=BaseLayerConfig)



from typing import Any, Dict

from .base import BaseSingleDomainRankingMethod
from .collectors import MCValueRankingDataCollector
from .rankers import MCValueRanker
from .rankingdata import MCValueRankingData


class MCValueMethod(BaseSingleDomainRankingMethod[MCValueRankingData]):
    def __init__(self) -> None:
        collector = MCValueRankingDataCollector()
        ranker = MCValueRanker()
        super().__init__(collector, ranker)

    @classmethod
    def collect_data_from_dict(cls, obj: Dict[str, Any]) -> MCValueRankingData:
        return MCValueRankingData(**obj)



from .combiner import FilterCombiner
from .term import (
    BaseCandidateTermFilter,
    EnglishConcatenationFilter,
    EnglishNumericFilter,
    EnglishProperNounFilter,
    EnglishSymbolLikeFilter,
    JapaneseConcatenationFilter,
    JapaneseNumericFilter,
    JapaneseProperNounFilter,
    JapaneseSymbolLikeFilter,
)
from .token import BaseCandidateTokenFilter, EnglishTokenFilter, JapaneseTokenFilter

# isort: unique-list
__all__ = [
    "BaseCandidateTermFilter",
    "BaseCandidateTokenFilter",
    "EnglishConcatenationFilter",
    "EnglishNumericFilter",
    "EnglishProperNounFilter",
    "EnglishSymbolLikeFilter",
    "EnglishTokenFilter",
    "FilterCombiner",
    "JapaneseConcatenationFilter",
    "JapaneseNumericFilter",
    "JapaneseProperNounFilter",
    "JapaneseSymbolLikeFilter",
    "JapaneseTokenFilter",
]



from .augmenter import AugmenterMapper
from .filter import CandidateTermFilterMapper, CandidateTokenFilterMapper
from .lang import LanguageTokenizerMapper
from .splitter import SplitterMapper
from .classifier import TokenClassifilerMapper

# isort: unique-list
__all__ = [
    "AugmenterMapper",
    "CandidateTermFilterMapper",
    "CandidateTokenFilterMapper",
    "LanguageTokenizerMapper",
    "SplitterMapper",
    "TokenClassifilerMapper",
]



import json
import os
from glob import glob
from shutil import rmtree
from typing import Any, Callable, Dict, List, Union

from py_pdf_term.methods import MethodTermRanking
from py_pdf_term.methods._methods.rankingdata import RankingData

from ...configs import MethodLayerConfig
from ..util import create_dir_name_from_config, create_file_name_from_paths
from .base import BaseMethodLayerDataCache, BaseMethodLayerRankingCache


class MethodLayerRankingFileCache(BaseMethodLayerRankingCache):
    def __init__(self, cache_dir: str) -> None:
        super().__init__(cache_dir)
        self._cache_dir = cache_dir

    def load(
        self,
        pdf_paths: List[str],
        config: MethodLayerConfig,
    ) -> Union[MethodTermRanking, None]:
        dir_name = create_dir_name_from_config(config, prefix="rank")
        file_name = create_file_name_from_paths(pdf_paths, "json")
        cache_file_path = os.path.join(self._cache_dir, dir_name, file_name)

        if not os.path.isfile(cache_file_path):
            return None

        with open(cache_file_path, "r") as json_file:
            try:
                obj = json.load(json_file)
            except json.JSONDecodeError:
                return None

        return MethodTermRanking.from_dict(obj)

    def store(
        self,
        pdf_paths: List[str],
        term_ranking: MethodTermRanking,
        config: MethodLayerConfig,
    ) -> None:
        dir_name = create_dir_name_from_config(config, prefix="rank")
        file_name = create_file_name_from_paths(pdf_paths, "json")
        cache_file_path = os.path.join(self._cache_dir, dir_name, file_name)

        os.makedirs(os.path.dirname(cache_file_path), exist_ok=True)

        with open(cache_file_path, "w") as json_file:
            json.dump(term_ranking.to_dict(), json_file, ensure_ascii=False, indent=2)

    def remove(self, pdf_paths: List[str], config: MethodLayerConfig) -> None:
        dir_name = create_dir_name_from_config(config, prefix="rank")
        file_name = create_file_name_from_paths(pdf_paths, "json")
        cache_dir_path = os.path.join(self._cache_dir, dir_name)
        cache_file_path = os.path.join(cache_dir_path, file_name)

        if not os.path.isfile(cache_file_path):
            return

        os.remove(cache_file_path)

        cache_file_paths = glob(os.path.join(cache_dir_path, "*.json"))
        if not cache_file_paths:
            rmtree(cache_dir_path)


class MethodLayerDataFileCache(BaseMethodLayerDataCache[RankingData]):
    def __init__(self, cache_dir: str) -> None:
        super().__init__(cache_dir)
        self._cache_dir = cache_dir

    def load(
        self,
        pdf_paths: List[str],
        config: MethodLayerConfig,
        from_dict: Callable[[Dict[str, Any]], RankingData],
    ) -> Union[RankingData, None]:
        dir_name = create_dir_name_from_config(config, prefix="data")
        file_name = create_file_name_from_paths(pdf_paths, "json")
        cache_file_path = os.path.join(self._cache_dir, dir_name, file_name)

        if not os.path.isfile(cache_file_path):
            return None

        with open(cache_file_path, "r") as json_file:
            obj = json.load(json_file)

        return from_dict(obj)

    def store(
        self,
        pdf_paths: List[str],
        ranking_data: RankingData,
        config: MethodLayerConfig,
    ) -> None:
        dir_name = create_dir_name_from_config(config, prefix="data")
        file_name = create_file_name_from_paths(pdf_paths, "json")
        cache_file_path = os.path.join(self._cache_dir, dir_name, file_name)

        os.makedirs(os.path.dirname(cache_file_path), exist_ok=True)

        with open(cache_file_path, "w") as json_file:
            json.dump(ranking_data.to_dict(), json_file, ensure_ascii=False, indent=2)

    def remove(self, pdf_paths: List[str], config: MethodLayerConfig) -> None:
        dir_name = create_dir_name_from_config(config, prefix="data")
        file_name = create_file_name_from_paths(pdf_paths, "json")
        cache_dir_path = os.path.join(self._cache_dir, dir_name)
        cache_file_path = os.path.join(cache_dir_path, file_name)

        if not os.path.isfile(cache_file_path):
            return

        os.remove(cache_file_path)

        cache_file_paths = glob(os.path.join(cache_dir_path, "*.json"))
        if not cache_file_paths:
            rmtree(cache_dir_path)



from abc import ABCMeta, abstractmethod
from typing import Generic, List

from py_pdf_term.candidates import DomainCandidateTermList

from ..data import MethodTermRanking
from ..rankingdata.base import RankingData


class BaseSingleDomainRanker(Generic[RankingData], metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def rank_terms(
        self,
        domain_candidates: DomainCandidateTermList,
        ranking_data: RankingData,
    ) -> MethodTermRanking:
        raise NotImplementedError(f"{self.__class__.__name__}.rank_terms()")


class BaseMultiDomainRanker(Generic[RankingData], metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def rank_terms(
        self,
        domain_candidates: DomainCandidateTermList,
        ranking_data_list: List[RankingData],
    ) -> MethodTermRanking:
        raise NotImplementedError(f"{self.__class__.__name__}.rank_terms()")



from abc import ABCMeta, abstractmethod
from typing import List, Optional

from py_pdf_term._common.data import Term
from ..classifiers import (
    BaseTokenClassifier,
    EnglishTokenClassifier,
    JapaneseTokenClassifier,
)


class BaseSplitter(metaclass=ABCMeta):
    def __init__(self, classifiers: Optional[List[BaseTokenClassifier]] = None) -> None:
        if classifiers is None:
            classifiers = [
                JapaneseTokenClassifier(),
                EnglishTokenClassifier(),
            ]

        self._classifiers = classifiers

    @abstractmethod
    def split(self, term: Term) -> List[Term]:
        raise NotImplementedError(f"{self.__class__.__name__}.split()")



from .base import BaseCandidateLayerCache
from .file import CandidateLayerFileCache
from .nocache import CandidateLayerNoCache

# isort: unique-list
__all__ = [
    "BaseCandidateLayerCache",
    "CandidateLayerFileCache",
    "CandidateLayerNoCache",
]



from .base import BaseCandidateTermFilter
from .concatenation import EnglishConcatenationFilter, JapaneseConcatenationFilter
from .numeric import EnglishNumericFilter, JapaneseNumericFilter
from .propernoun import EnglishProperNounFilter, JapaneseProperNounFilter
from .symbollike import EnglishSymbolLikeFilter, JapaneseSymbolLikeFilter

# isort: unique-list
__all__ = [
    "BaseCandidateTermFilter",
    "EnglishConcatenationFilter",
    "EnglishNumericFilter",
    "EnglishProperNounFilter",
    "EnglishSymbolLikeFilter",
    "JapaneseConcatenationFilter",
    "JapaneseNumericFilter",
    "JapaneseProperNounFilter",
    "JapaneseSymbolLikeFilter",
]



from typing import Type

from ...caches import BaseStylingLayerCache, StylingLayerFileCache, StylingLayerNoCache
from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class StylingLayerCacheMapper(BaseMapper[Type[BaseStylingLayerCache]]):
    @classmethod
    def default_mapper(cls) -> "StylingLayerCacheMapper":
        default_mapper = cls()

        cache_clses = [StylingLayerNoCache, StylingLayerFileCache]
        for cache_cls in cache_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{cache_cls.__name__}", cache_cls)

        return default_mapper



from .base import BaseRankingData, RankingData
from .flr import FLRRankingData
from .flrh import FLRHRankingData
from .hits import HITSRankingData
from .mcvalue import MCValueRankingData
from .mdp import MDPRankingData
from .tfidf import TFIDFRankingData

# isort: unique-list
__all__ = [
    "BaseRankingData",
    "FLRHRankingData",
    "FLRRankingData",
    "HITSRankingData",
    "MCValueRankingData",
    "MDPRankingData",
    "RankingData",
    "TFIDFRankingData",
]



from typing import Type

from py_pdf_term.candidates.augmenters import (
    BaseAugmenter,
    EnglishConnectorTermAugmenter,
    JapaneseConnectorTermAugmenter,
)

from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class AugmenterMapper(BaseMapper[Type[BaseAugmenter]]):
    @classmethod
    def default_mapper(cls) -> "AugmenterMapper":
        default_mapper = cls()

        augmenter_clses = [
            JapaneseConnectorTermAugmenter,
            EnglishConnectorTermAugmenter,
        ]
        for augmenter_cls in augmenter_clses:
            default_mapper.add(
                f"{PACKAGE_NAME}.{augmenter_cls.__name__}", augmenter_cls
            )

        return default_mapper



from py_pdf_term.analysis import ContainerTermsAnalyzer, TermOccurrenceAnalyzer
from py_pdf_term.candidates import DomainCandidateTermList

from ..rankingdata import MCValueRankingData
from .base import BaseRankingDataCollector


class MCValueRankingDataCollector(BaseRankingDataCollector[MCValueRankingData]):
    def __init__(self) -> None:
        super().__init__()
        self._termocc_analyzer = TermOccurrenceAnalyzer()
        self._containers_analyzer = ContainerTermsAnalyzer()

    def collect(self, domain_candidates: DomainCandidateTermList) -> MCValueRankingData:
        termocc = self._termocc_analyzer.analyze(domain_candidates)
        container_terms = self._containers_analyzer.analyze(domain_candidates)

        return MCValueRankingData(
            domain_candidates.domain,
            termocc.term_freq,
            container_terms.container_terms,
        )



from .concats import DomainLeftRightFrequency, TermLeftRightFrequencyAnalyzer
from .cooccurrences import ContainerTermsAnalyzer, DomainContainerTerms
from .occurrences import DomainTermOccurrence, TermOccurrenceAnalyzer

# isort: unique-list
__all__ = [
    "ContainerTermsAnalyzer",
    "DomainContainerTerms",
    "DomainLeftRightFrequency",
    "DomainTermOccurrence",
    "TermLeftRightFrequencyAnalyzer",
    "TermOccurrenceAnalyzer",
]



from typing import Callable, Iterable, List

from py_pdf_term._common.data import ScoredTerm, Term
from py_pdf_term._common.extended_math import extended_log10
from py_pdf_term.candidates import DomainCandidateTermList

from ..data import MethodTermRanking
from ..rankingdata import MDPRankingData
from .base import BaseMultiDomainRanker


class MDPRanker(BaseMultiDomainRanker[MDPRankingData]):
    def __init__(
        self, compile_scores: Callable[[Iterable[float]], float] = min
    ) -> None:
        self._compile_scores = compile_scores

    def rank_terms(
        self,
        domain_candidates: DomainCandidateTermList,
        ranking_data_list: List[MDPRankingData],
    ) -> MethodTermRanking:
        domain_candidates_dict = domain_candidates.to_nostyle_candidates_dict(
            to_str=lambda candidate: candidate.lemma()
        )
        ranking_data = next(
            filter(
                lambda item: item.domain == domain_candidates.domain,
                ranking_data_list,
            )
        )
        other_ranking_data_list = list(
            filter(
                lambda item: item.domain != domain_candidates.domain,
                ranking_data_list,
            )
        )
        ranking = list(
            map(
                lambda candidate: self._calculate_score(
                    candidate, ranking_data, other_ranking_data_list
                ),
                domain_candidates_dict.values(),
            )
        )
        ranking.sort(key=lambda term: -term.score)
        return MethodTermRanking(domain_candidates.domain, ranking)

    def _calculate_score(
        self,
        candidate: Term,
        ranking_data: MDPRankingData,
        other_ranking_data_list: List[MDPRankingData],
    ) -> ScoredTerm:
        candidate_lemma = candidate.lemma()
        score = self._compile_scores(
            map(
                lambda other_ranking_data: self._calculate_zvalue(
                    candidate, ranking_data, other_ranking_data
                ),
                other_ranking_data_list,
            )
        )

        return ScoredTerm(candidate_lemma, score)

    def _calculate_zvalue(
        self,
        candidate: Term,
        our_ranking_data: MDPRankingData,
        their_ranking_data: MDPRankingData,
    ) -> float:
        candidate_lemma = candidate.lemma()

        our_term_freq = our_ranking_data.term_freq.get(candidate_lemma, 0)
        their_term_freq = their_ranking_data.term_freq.get(candidate_lemma, 0)

        our_inum_terms = 1 / our_ranking_data.num_terms
        their_inum_terms = 1 / their_ranking_data.num_terms

        our_term_prob = our_term_freq / our_ranking_data.num_terms
        their_term_prob = their_term_freq / their_ranking_data.num_terms

        term_prob = (our_term_freq + their_term_freq) / (
            our_ranking_data.num_terms + their_ranking_data.num_terms
        )

        return extended_log10(
            (our_term_prob - their_term_prob)
            / (term_prob * (1.0 - term_prob) * (our_inum_terms + their_inum_terms))
        )



from abc import ABCMeta, abstractmethod
from py_pdf_term.tokenizer import Token


class BaseTokenClassifier(metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def inscope(self, token: Token) -> bool:
        raise NotImplementedError(f"{self.__class__.__name__}.inscope()")

    @abstractmethod
    def is_symbol(self, token: Token) -> bool:
        raise NotImplementedError(f"{self.__class__.__name__}.is_symbol()")

    @abstractmethod
    def is_connector_symbol(self, token: Token) -> bool:
        raise NotImplementedError(f"{self.__class__.__name__}.is_connector_symbol()")

    @abstractmethod
    def is_connector_term(self, token: Token) -> bool:
        raise NotImplementedError(f"{self.__class__.__name__}.is_connector_term()")

    def is_meaningless(self, token: Token) -> bool:
        return (
            self.is_symbol(token)
            or self.is_connector_symbol(token)
            or self.is_connector_term(token)
        )

    def is_connector(self, token: Token) -> bool:
        return self.is_connector_symbol(token) or self.is_connector_term(token)



import re
from typing import Optional
from unicodedata import normalize

from py_pdf_term._common.consts import FULLWIDTH_ASCII_CHARS, HALFWIDTH_ASCII_CHARS

ERROR = re.compile(r"[\x00-\x08\x0b-\x0c\x0e-\x1f]|\(cid:\d+\)")
SPACES = re.compile(r"\s+")
ASCII_FULL2HALF_TABLE = str.maketrans(FULLWIDTH_ASCII_CHARS, HALFWIDTH_ASCII_CHARS)


def clean_content_text(
    text: str,
    nfc_norm: bool,
    include_pattern: Optional[str],
    exclude_pattern: Optional[str],
) -> str:
    text = ERROR.sub("", text)
    text = SPACES.sub(" ", text).strip()
    text = text.translate(ASCII_FULL2HALF_TABLE)

    if nfc_norm:
        text = normalize("NFC", text)

    return (
        text
        if (include_pattern is None or re.search(include_pattern, text) is not None)
        and (exclude_pattern is None or re.search(exclude_pattern, text) is None)
        else ""
    )



from abc import ABCMeta, abstractmethod
from typing import List

from .data import Token


class BaseLanguageTokenizer(metaclass=ABCMeta):
    @abstractmethod
    def inscope(self, text: str) -> bool:
        raise NotImplementedError(f"{self.__class__.__name__}.inscope()")

    @abstractmethod
    def tokenize(self, text: str) -> List[Token]:
        raise NotImplementedError(f"{self.__class__.__name__}.tokenize()")



from .data import DomainStylingScoreList, PageStylingScoreList, PDFStylingScoreList
from .scorer import StylingScorer

# isort: unique-list
__all__ = [
    "DomainStylingScoreList",
    "PDFStylingScoreList",
    "PageStylingScoreList",
    "StylingScorer",
]



from typing import Union

from py_pdf_term.candidates import PDFCandidateTermList

from ...configs import CandidateLayerConfig
from .base import BaseCandidateLayerCache


class CandidateLayerNoCache(BaseCandidateLayerCache):
    def __init__(self, cache_dir: str) -> None:
        pass

    def load(
        self, pdf_path: str, config: CandidateLayerConfig
    ) -> Union[PDFCandidateTermList, None]:
        pass

    def store(
        self, candidates: PDFCandidateTermList, config: CandidateLayerConfig
    ) -> None:
        pass

    def remove(self, pdf_path: str, config: CandidateLayerConfig) -> None:
        pass



from dataclasses import dataclass
from typing import Dict

from py_pdf_term._common.data import Term
from py_pdf_term.candidates import DomainCandidateTermList


from ..runner import AnalysisRunner


@dataclass(frozen=True)
class DomainLeftRightFrequency:
    domain: str
    # unique domain name
    left_freq: Dict[str, Dict[str, int]]
    # number of occurrences of lemmatized (left, token) in the domain
    # if token or left is meaningless, this is fixed at zero
    right_freq: Dict[str, Dict[str, int]]
    # number of occurrences of lemmatized (token, right) in the domain
    # if token or right is meaningless, this is fixed at zero


class TermLeftRightFrequencyAnalyzer:
    def __init__(self, ignore_augmented: bool = True) -> None:
        self._ignore_augmented = ignore_augmented
        self._runner = AnalysisRunner(ignore_augmented=ignore_augmented)

    def analyze(
        self, domain_candidates: DomainCandidateTermList
    ) -> DomainLeftRightFrequency:
        def update(
            lrfreq: DomainLeftRightFrequency,
            pdf_id: int,
            page_num: int,
            candidate: Term,
        ) -> None:
            num_tokens = len(candidate.tokens)
            for i in range(num_tokens):
                token = candidate.tokens[i]
                if token.is_meaningless:
                    lrfreq.left_freq[token.lemma] = dict()
                    lrfreq.right_freq[token.lemma] = dict()
                    continue

                self._update_left_freq(lrfreq, candidate, i)
                self._update_right_freq(lrfreq, candidate, i)

        lrfreq = self._runner.run_through_candidates(
            domain_candidates,
            DomainLeftRightFrequency(domain_candidates.domain, dict(), dict()),
            update,
        )

        return lrfreq

    def _update_left_freq(
        self, lrfreq: DomainLeftRightFrequency, candidate: Term, idx: int
    ) -> None:
        token = candidate.tokens[idx]

        if idx == 0:
            left = lrfreq.left_freq.get(token.lemma, dict())
            lrfreq.left_freq[token.lemma] = left
            return

        left_token = candidate.tokens[idx - 1]
        if not left_token.is_meaningless:
            left = lrfreq.left_freq.get(token.lemma, dict())
            left[left_token.lemma] = left.get(left_token.lemma, 0) + 1
            lrfreq.left_freq[token.lemma] = left
        else:
            left = lrfreq.left_freq.get(token.lemma, dict())
            lrfreq.left_freq[token.lemma] = left

    def _update_right_freq(
        self, lrfreq: DomainLeftRightFrequency, candidate: Term, idx: int
    ) -> None:
        num_tokens = len(candidate.tokens)
        token = candidate.tokens[idx]

        if idx == num_tokens - 1:
            right = lrfreq.right_freq.get(token.lemma, dict())
            lrfreq.right_freq[token.lemma] = right
            return

        right_token = candidate.tokens[idx + 1]
        if not right_token.is_meaningless:
            right = lrfreq.right_freq.get(token.lemma, dict())
            right[right_token.lemma] = right.get(right_token.lemma, 0) + 1
            lrfreq.right_freq[token.lemma] = right
        else:
            right = lrfreq.right_freq.get(token.lemma, dict())
            lrfreq.right_freq[token.lemma] = right



from py_pdf_term._common.data import Term
from py_pdf_term.tokenizer import Token
from ....classifiers import JapaneseTokenClassifier

from ..base import BaseJapaneseCandidateTermFilter


class JapaneseNumericFilter(BaseJapaneseCandidateTermFilter):
    def __init__(self) -> None:
        self._classifier = JapaneseTokenClassifier()

    def is_candidate(self, scoped_term: Term) -> bool:
        return not self._is_numeric_phrase(scoped_term)

    def _is_numeric_phrase(self, scoped_term: Term) -> bool:
        def is_numeric_or_meaningless(token: Token) -> bool:
            return (
                token.pos == "接頭辞"
                or (token.pos == "名詞" and token.category == "数詞")
                or (
                    token.pos == "名詞"
                    and token.category == "普通名詞"
                    and token.subcategory == "助数詞可能"
                )
                or token.pos == "接尾辞"
                or self._classifier.is_meaningless(token)
            )

        return all(map(is_numeric_or_meaningless, scoped_term.tokens))



from .base import BaseRankingDataCollector
from .flr import FLRRankingDataCollector
from .flrh import FLRHRankingDataCollector
from .hits import HITSRankingDataCollector
from .mcvalue import MCValueRankingDataCollector
from .mdp import MDPRankingDataCollector
from .tfidf import TFIDFRankingDataCollector

# isort: unique-list
__all__ = [
    "BaseRankingDataCollector",
    "FLRHRankingDataCollector",
    "FLRRankingDataCollector",
    "HITSRankingDataCollector",
    "MCValueRankingDataCollector",
    "MDPRankingDataCollector",
    "TFIDFRankingDataCollector",
]



from typing import List, Optional

from py_pdf_term._common.data import Term
from py_pdf_term.tokenizer import Token

from .term import (
    BaseCandidateTermFilter,
    EnglishConcatenationFilter,
    EnglishNumericFilter,
    EnglishProperNounFilter,
    EnglishSymbolLikeFilter,
    JapaneseConcatenationFilter,
    JapaneseNumericFilter,
    JapaneseProperNounFilter,
    JapaneseSymbolLikeFilter,
)
from .token import BaseCandidateTokenFilter, EnglishTokenFilter, JapaneseTokenFilter


class FilterCombiner:
    def __init__(
        self,
        token_filters: Optional[List[BaseCandidateTokenFilter]] = None,
        term_filters: Optional[List[BaseCandidateTermFilter]] = None,
    ) -> None:
        if token_filters is None:
            token_filters = [
                JapaneseTokenFilter(),
                EnglishTokenFilter(),
            ]
        if term_filters is None:
            term_filters = [
                JapaneseConcatenationFilter(),
                EnglishConcatenationFilter(),
                JapaneseSymbolLikeFilter(),
                EnglishSymbolLikeFilter(),
                JapaneseProperNounFilter(),
                EnglishProperNounFilter(),
                JapaneseNumericFilter(),
                EnglishNumericFilter(),
            ]

        self._token_filters = token_filters
        self._term_filters = term_filters

    def is_partof_candidate(self, tokens: List[Token], idx: int) -> bool:
        token = tokens[idx]
        if all(map(lambda mf: not mf.inscope(token), self._token_filters)):
            return False

        return all(
            map(
                lambda mf: not mf.inscope(token) or mf.is_partof_candidate(tokens, idx),
                self._token_filters,
            )
        )

    def is_candidate(self, term: Term) -> bool:
        if all(map(lambda tf: not tf.inscope(term), self._term_filters)):
            return False

        return all(
            map(
                lambda tf: not tf.inscope(term) or tf.is_candidate(term),
                self._term_filters,
            )
        )



from dataclasses import dataclass
from math import sqrt
from typing import Dict

from py_pdf_term._common.data import ScoredTerm, Term
from py_pdf_term._common.extended_math import extended_log10
from py_pdf_term.candidates import DomainCandidateTermList

from ..data import MethodTermRanking
from ..rankingdata import HITSRankingData
from .base import BaseSingleDomainRanker


@dataclass(frozen=True)
class HITSAuthHubData:
    token_auth: Dict[str, float]
    # auth value of the token
    # the more tokens links to, the larger the auth value becomes
    # initial auth value is 1.0
    token_hub: Dict[str, float]
    # hub value of the term
    # the more tokens is linked from, the larger the hub value becomes
    # initial hub value is 1.0


class HITSRanker(BaseSingleDomainRanker[HITSRankingData]):
    def __init__(self, threshold: float = 1e-8, max_loop: int = 1000) -> None:
        self._threshold = threshold
        self._max_loop = max_loop

    def rank_terms(
        self, domain_candidates: DomainCandidateTermList, ranking_data: HITSRankingData
    ) -> MethodTermRanking:
        auth_hub_data = self._create_auth_hub_data(ranking_data)
        domain_candidates_dict = domain_candidates.to_nostyle_candidates_dict(
            to_str=lambda candidate: candidate.lemma()
        )
        ranking = list(
            map(
                lambda candidate: self._calculate_score(
                    candidate, ranking_data, auth_hub_data
                ),
                domain_candidates_dict.values(),
            )
        )
        ranking.sort(key=lambda term: -term.score)
        return MethodTermRanking(domain_candidates.domain, ranking)

    def _create_auth_hub_data(self, ranking_data: HITSRankingData) -> HITSAuthHubData:
        token_auth: Dict[str, float] = {
            token_lemma: 1.0 for token_lemma in ranking_data.left_freq
        }
        token_hub: Dict[str, float] = {
            token_lemma: 1.0 for token_lemma in ranking_data.right_freq
        }

        converged = False
        loop = 0
        while not converged and loop < self._max_loop:
            new_token_auth = {
                token: sum(map(lambda hub: token_hub[hub], left.keys()), 0.0)
                for token, left in ranking_data.left_freq.items()
            }
            auth_norm = sqrt(sum(map(lambda x: x * x, new_token_auth.values())))
            new_token_auth = {
                token: auth_score / auth_norm
                for token, auth_score in new_token_auth.items()
            }

            new_token_hub = {
                token: sum(map(lambda auth: token_auth[auth], right.keys()), 0.0)
                for token, right in ranking_data.right_freq.items()
            }
            hub_norm = sqrt(sum(map(lambda x: x * x, new_token_hub.values())))
            new_token_hub = {
                token: hub_score / hub_norm
                for token, hub_score in new_token_hub.items()
            }

            converged = all(
                [
                    abs(new_token_auth[token] - token_auth[token]) < self._threshold
                    for token in ranking_data.left_freq
                ]
                + [
                    abs(new_token_hub[token] - token_hub[token]) < self._threshold
                    for token in ranking_data.right_freq
                ]
            )

            token_auth = new_token_auth
            token_hub = new_token_hub

            loop += 1

        return HITSAuthHubData(token_auth, token_hub)

    def _calculate_score(
        self,
        candidate: Term,
        ranking_data: HITSRankingData,
        auth_hub_data: HITSAuthHubData,
    ) -> ScoredTerm:
        candidate_lemma = candidate.lemma()
        num_tokens = len(candidate.tokens)
        num_meaningless_tokens = sum(
            map(lambda token: 1 if token.is_meaningless else 0, candidate.tokens)
        )

        if num_tokens == 0:
            return ScoredTerm(candidate_lemma, 0.0)

        term_freq_score = extended_log10(ranking_data.term_freq.get(candidate_lemma, 0))

        if num_tokens == 1:
            token_lemma = candidate.tokens[0].lemma
            auth_hub_score = 0.5 * (
                extended_log10(auth_hub_data.token_hub.get(token_lemma, 0.0))
                + extended_log10(auth_hub_data.token_auth.get(token_lemma, 0.0))
            )
            score = term_freq_score + auth_hub_score
            return ScoredTerm(candidate_lemma, score)

        auth_hub_score = 0.0
        for i, token in enumerate(candidate.tokens):
            if token.is_meaningless:
                continue

            if i == 0:
                auth_hub_score += extended_log10(
                    auth_hub_data.token_hub.get(token.lemma, 0.0)
                )
            elif i == num_tokens - 1:
                auth_hub_score += extended_log10(
                    auth_hub_data.token_auth.get(token.lemma, 0.0)
                )
            else:
                auth_hub_score += extended_log10(
                    auth_hub_data.token_hub.get(token.lemma, 0.0)
                )
                auth_hub_score += extended_log10(
                    auth_hub_data.token_auth.get(token.lemma, 0.0)
                )
                auth_hub_score = auth_hub_score / 2

        auth_hub_score /= num_tokens - num_meaningless_tokens

        score = term_freq_score + auth_hub_score
        return ScoredTerm(candidate_lemma, score)



from typing import List, Optional

from py_pdf_term.techterms import PDFTechnicalTermList

from .caches import DEFAULT_CACHE_DIR
from .configs import (
    CandidateLayerConfig,
    MethodLayerConfig,
    StylingLayerConfig,
    TechnicalTermLayerConfig,
    XMLLayerConfig,
)
from .data import DomainPDFList
from .layers import (
    CandidateLayer,
    MethodLayer,
    StylingLayer,
    TechnicalTermLayer,
    XMLLayer,
)
from .mappers import (
    AugmenterMapper,
    BinaryOpenerMapper,
    CandidateLayerCacheMapper,
    CandidateTermFilterMapper,
    CandidateTokenFilterMapper,
    LanguageTokenizerMapper,
    MethodLayerDataCacheMapper,
    MethodLayerRankingCacheMapper,
    MultiDomainRankingMethodMapper,
    SingleDomainRankingMethodMapper,
    SplitterMapper,
    StylingLayerCacheMapper,
    StylingScoreMapper,
    TokenClassifilerMapper,
    XMLLayerCacheMapper,
)


class PyPDFTermExtractor:
    """Top level class of py-pdf-term. E2E class to extract terminologies from PDF file.

    Args:
        xml_config:
            Config of XML Layer.

        candidate_config:
            Config of Candidate Term Layer.

        method_config:
            Config of Method Layer.

        styling_config:
            Config of Styling Layer.

        techterm_config:
            Config of Technial Term Layer.

        bin_opener_mapper:
            Mapper from `xml_config.open_bin` to a function to open a input PDF file in
            the binary mode. This is used in XML Layer.

        lang_tokenizer_mapper:
            Mapper from an element in `candidate_config.lang_tokenizers` to a class to
            tokenize texts in a specific language with spaCy. This is used in Candidate
            Term Layer.

        token_classifier_mapper:
            Mapper from an element in `candidate_config.token_classifiers` to a class to
            classify tokens into True/False by several functions. This is used in
            Candidate Term Layer.

        token_filter_mapper:
            Mapper from an element in `candidate_config.token_filters` to a class to
            filter tokens which are likely to be parts of candidates. This is used in
            Candidate Term Layer.

        term_filter_mapper:
            Mapper from an element in `candidate_config.term_filters` to a class to
            filter terms which are likely to be candidates. This is used in Candidate
            Term Layer.

        splitter_mapper:
            Mapper from an element in `candidate_config.splitters` to a class to split
            too long terms or wrongly concatenated terms. When XML Layer extracts a text
            from a table, figure or something in a PDF file, texts in them are often
            wrongly concatenated so that the text looks like one too long candidate
            term. This is used in Candidate Term Layer.
            e.g. If a PDF have a table to compare sort algorithms and the header text
            wanna mean "Selection Sort" | "Quick Sort" | "Merge Sort", the text is
            recognized as a word like "Selection Sort Quick Sort Merge Sort". This
            wrongly concatenated word must be split down into three algorithm names.

        augmenter_mapper:
            Mapper from an element in `candidate_config.augmenters` to a class to
            augment candidates. The augumentation means that if a long candidate term is
            found, sub-terms of the original term could also be candidates. This is used
            in Candidate Term Layer.
            e.g. Original long term: "Structure of Layered Architecture". Augmented sub-
            terms: "Structure" and "Layered Architecture".

        single_method_mapper:
            Mapper from `method_config.method` to a class to calculate method scores of
            candidate terms when `method_config.method_type` equals to "single".
            A single-domain ranking method is an ranking algorithm which can calculate
            method scores without cross-domain information.
            e.g. Term-Frequency algorithm uses only frequency of occurcences of terms.
            Therefore, TF is a single-domain ranking method.

        multi_method_mapper:
            Mapper from `method_config.method` to a class to calculate method scores of
            candidate terms when `method_config.method_type` equals to "multi".
            A multi-domain ranking method is an ranking algorithm which calculate method
            scores with cross-domain information.
            e.g. TF-IDF algorithm uses Inverse Document Frequency, which goes to high
            then a term appears frequently in a target domain but not in the other
            domains. To find IDF score, cross-domain data are required. Therefore,
            TF-IDF is a multi-domain ranking method.

        styling_score_mapper:
            Mapper from an element in `styling_config.styling_scores` to a class to
            calculate styling scores of candidate terms. Each scoring class is expected
            to focus one specific styling feature. Then different types of scores are
            combined later.

        xml_cache_mapper:
            Mapper from `xml_config.cache` to a class to provide XML Layer with the
            cache  mechanism. The xml cache manages XML files converted from input PDF
            files.

        candidate_cache_mapper:
            Mapper from `candidate_config.cache` to a class to provide Candidate Term
            Layer with the cache mechanism. The candidate cache manages lists of
            candidate terms.

        method_ranking_cache_mapper:
            Mapper from `method_config.ranking_cache` to a class to provide Method Layer
            with the cache mechanism. The method ranking cache manages candidate terms
            ordered by the method scores.

        method_data_cache_mapper:
            Mapper from `method_config.data_cache` to a class to provide Method Layer
            with the cache mechanism. The method data cache manages analysis data of the
            candidate terms such as frequency or likelihood.

        styling_cache_mapper:
            Mapper from `styling_config.cache` to a class to provide Styling Layer with
            the cache mechanism. The styling cache manages candidate terms ordered by
            the styling scores.

        cache_dir:
            Path like string where cache files to be stored. For example, path to a
            local dirctory, a url or a bucket name of a cloud storage service.
    """

    def __init__(
        self,
        xml_config: Optional[XMLLayerConfig] = None,
        candidate_config: Optional[CandidateLayerConfig] = None,
        method_config: Optional[MethodLayerConfig] = None,
        styling_config: Optional[StylingLayerConfig] = None,
        techterm_config: Optional[TechnicalTermLayerConfig] = None,
        bin_opener_mapper: Optional[BinaryOpenerMapper] = None,
        lang_tokenizer_mapper: Optional[LanguageTokenizerMapper] = None,
        token_classifier_mapper: Optional[TokenClassifilerMapper] = None,
        token_filter_mapper: Optional[CandidateTokenFilterMapper] = None,
        term_filter_mapper: Optional[CandidateTermFilterMapper] = None,
        splitter_mapper: Optional[SplitterMapper] = None,
        augmenter_mapper: Optional[AugmenterMapper] = None,
        single_method_mapper: Optional[SingleDomainRankingMethodMapper] = None,
        multi_method_mapper: Optional[MultiDomainRankingMethodMapper] = None,
        styling_score_mapper: Optional[StylingScoreMapper] = None,
        xml_cache_mapper: Optional[XMLLayerCacheMapper] = None,
        candidate_cache_mapper: Optional[CandidateLayerCacheMapper] = None,
        method_ranking_cache_mapper: Optional[MethodLayerRankingCacheMapper] = None,
        method_data_cache_mapper: Optional[MethodLayerDataCacheMapper] = None,
        styling_cache_mapper: Optional[StylingLayerCacheMapper] = None,
        cache_dir: str = DEFAULT_CACHE_DIR,
    ) -> None:
        xml_layer = XMLLayer(
            config=xml_config,
            bin_opener_mapper=bin_opener_mapper,
            cache_mapper=xml_cache_mapper,
            cache_dir=cache_dir,
        )
        candidate_layer = CandidateLayer(
            xml_layer=xml_layer,
            config=candidate_config,
            lang_tokenizer_mapper=lang_tokenizer_mapper,
            token_classifier_mapper=token_classifier_mapper,
            token_filter_mapper=token_filter_mapper,
            term_filter_mapper=term_filter_mapper,
            splitter_mapper=splitter_mapper,
            augmenter_mapper=augmenter_mapper,
            cache_mapper=candidate_cache_mapper,
            cache_dir=cache_dir,
        )
        method_layer = MethodLayer(
            candidate_layer=candidate_layer,
            config=method_config,
            single_method_mapper=single_method_mapper,
            multi_method_mapper=multi_method_mapper,
            ranking_cache_mapper=method_ranking_cache_mapper,
            data_cache_mapper=method_data_cache_mapper,
            cache_dir=cache_dir,
        )
        styling_layer = StylingLayer(
            candidate_layer=candidate_layer,
            config=styling_config,
            styling_score_mapper=styling_score_mapper,
            cache_mapper=styling_cache_mapper,
            cache_dir=cache_dir,
        )
        techterm_layer = TechnicalTermLayer(
            candidate_layer=candidate_layer,
            method_layer=method_layer,
            styling_layer=styling_layer,
            config=techterm_config,
        )

        self._techterm_layer = techterm_layer

    def extract(
        self,
        domain: str,
        pdf_path: str,
        single_domain_pdfs: Optional[DomainPDFList] = None,
        multi_domain_pdfs: Optional[List[DomainPDFList]] = None,
    ) -> PDFTechnicalTermList:
        """Function to extract terminologies from PDF file.

        Args:
            domain:
                Domain name which the input PDF file belongs to. This may be the name of
                a course, the name of a technical field or something.

            pdf_path:
                Path like string to the input PDF file which terminologies to be
                extracted. The file MUST belong to `domain`.

            single_domain_pdfs:
                List of path like strings to the PDF files which belong to `domain`.
                `single_domain_pdfs.domain` MUST equals to `domain`. This argument is
                required when a single-domain ranking method is to be used.

            multi_domain_pdfs:
                List of path like strings to the PDF files which classified by domain.
                There MUST be an element in `multi_domain_pdfs` whose domain equals to
                `domain`. This argument is required when a multi-domain ranking method
                is to be used.

        Returns:
            PDFTechnicalTermList:
                Terminology list per page from the input PDF file.
        """
        pdf_techterms = self._techterm_layer.create_pdf_techterms(
            domain, pdf_path, single_domain_pdfs, multi_domain_pdfs
        )
        return pdf_techterms



from py_pdf_term.analysis import TermLeftRightFrequencyAnalyzer, TermOccurrenceAnalyzer
from py_pdf_term.candidates import DomainCandidateTermList

from ..rankingdata import FLRHRankingData
from .base import BaseRankingDataCollector


class FLRHRankingDataCollector(BaseRankingDataCollector[FLRHRankingData]):
    def __init__(self) -> None:
        super().__init__()
        self._termocc_analyzer = TermOccurrenceAnalyzer()
        self._lrfreq_analyzer = TermLeftRightFrequencyAnalyzer()

    def collect(self, domain_candidates: DomainCandidateTermList) -> FLRHRankingData:
        termocc = self._termocc_analyzer.analyze(domain_candidates)
        lrfreq = self._lrfreq_analyzer.analyze(domain_candidates)

        return FLRHRankingData(
            domain_candidates.domain,
            termocc.term_freq,
            lrfreq.left_freq,
            lrfreq.right_freq,
        )



from typing import Callable, TypeVar

from py_pdf_term._common.data import Term
from py_pdf_term.candidates import DomainCandidateTermList

AnalysisResult = TypeVar("AnalysisResult")


class AnalysisRunner:
    def __init__(self, ignore_augmented: bool = True) -> None:
        self._ignore_augmented = ignore_augmented

    def run_through_candidates(
        self,
        domain_candidates: DomainCandidateTermList,
        initial_result: AnalysisResult,
        update_result: Callable[[AnalysisResult, int, int, Term], None],
    ) -> AnalysisResult:
        result = initial_result

        for pdf_id, pdf_candidates in enumerate(domain_candidates.pdfs):
            for page_candidates in pdf_candidates.pages:
                page_num = page_candidates.page_num
                for candidate in page_candidates.candidates:
                    if self._ignore_augmented and candidate.augmented:
                        continue
                    update_result(result, pdf_id, page_num, candidate)

        return result

    def run_through_subcandidates(
        self,
        domain_candidates: DomainCandidateTermList,
        initial_result: AnalysisResult,
        update_result: Callable[[AnalysisResult, int, int, Term], None],
    ) -> AnalysisResult:
        result = initial_result

        for pdf_id, pdf_candidates in enumerate(domain_candidates.pdfs):
            for page_candidates in pdf_candidates.pages:
                page_num = page_candidates.page_num
                for candidate in page_candidates.candidates:
                    if self._ignore_augmented and candidate.augmented:
                        continue

                    num_tokens = len(candidate.tokens)
                    for i in range(num_tokens):
                        for j in range(i + 1, num_tokens + 1):
                            subcandidate = Term(
                                candidate.tokens[i:j],
                                candidate.fontsize,
                                candidate.ncolor,
                                candidate.augmented,
                            )
                            update_result(result, pdf_id, page_num, subcandidate)

        return result



import re
from typing import List

from py_pdf_term._common.consts import ENGLISH_REGEX, JAPANESE_REGEX, NUMBER_REGEX
from py_pdf_term.tokenizer import Token

from .base import BaseCandidateTokenFilter


class JapaneseTokenFilter(BaseCandidateTokenFilter):
    def __init__(self) -> None:
        self._regex = re.compile(rf"({JAPANESE_REGEX}|{ENGLISH_REGEX}|{NUMBER_REGEX})+")

    def inscope(self, token: Token) -> bool:
        token_str = str(token)
        return token.lang == "ja" and (
            self._regex.fullmatch(token_str) is not None or token_str == "-"
        )

    def is_partof_candidate(self, tokens: List[Token], idx: int) -> bool:
        scoped_token = tokens[idx]
        num_tokens = len(tokens)

        if scoped_token.pos == "名詞":
            return (
                (
                    scoped_token.category == "普通名詞"
                    and scoped_token.subcategory
                    in {"一般", "サ変可能", "形状詞可能", "サ変形状詞可能", "助数詞可能"}
                )
                or scoped_token.category == "固有名詞"
                or scoped_token.category == "数詞"
            )
        elif scoped_token.pos in {"形状詞", "形容詞"}:
            return (
                scoped_token.category in {"一般"}
                and idx < num_tokens - 1
                and tokens[idx + 1].pos in {"名詞", "記号", "接尾辞", "形状詞", "形容詞"}
            )
        elif scoped_token.pos == "動詞":
            return (
                scoped_token.category in {"一般"}
                and idx < num_tokens - 1
                and tokens[idx + 1].pos in {"接尾辞", "動詞"}
            )
        elif scoped_token.pos == "接頭辞":
            return (
                idx < num_tokens - 1
                and tokens[idx + 1].pos in {"名詞", "記号", "形状詞"}
                # No line break
            )
        elif scoped_token.pos == "接尾辞":
            return (
                scoped_token.category in {"名詞的", "形状詞的"}
                and idx > 0
                and tokens[idx - 1].pos in {"名詞", "形状詞", "動詞", "形容詞", "記号"}
            )
        elif scoped_token.pos == "助詞":
            return scoped_token.surface_form == "の"
        elif scoped_token.pos == "記号":
            return self._regex.match(str(scoped_token)) is not None
        elif scoped_token.pos == "補助記号":
            scoped_token_str = str(scoped_token)
            if scoped_token_str not in {"-", "・"}:
                return False
            return (
                0 < idx < num_tokens - 1
                and self._regex.match(str(tokens[idx - 1])) is not None
                and self._regex.match(str(tokens[idx + 1])) is not None
            )

        return False



from py_pdf_term._common.data import Term

from ..base import BaseEnglishCandidateTermFilter


class EnglishProperNounFilter(BaseEnglishCandidateTermFilter):
    def __init__(self) -> None:
        pass

    def is_candidate(self, scoped_term: Term) -> bool:
        return True



from dataclasses import dataclass
from typing import Dict, Set

from py_pdf_term._common.data import Term
from py_pdf_term.candidates import DomainCandidateTermList

from ..runner import AnalysisRunner


@dataclass(frozen=True)
class DomainContainerTerms:
    domain: str
    # unique domain name
    container_terms: Dict[str, Set[str]]
    # set of lemmatized containers of the lemmatized term in the domain
    # (term, container) is valid iff the container contains the term
    # as a proper subsequence


class ContainerTermsAnalyzer:
    def __init__(self, ignore_augmented: bool = True) -> None:
        self._runner = AnalysisRunner(ignore_augmented=ignore_augmented)

    def analyze(
        self, domain_candidates: DomainCandidateTermList
    ) -> DomainContainerTerms:
        domain_candidates_set = domain_candidates.to_candidates_str_set(
            lambda candidate: candidate.lemma()
        )

        def update(
            container_terms: DomainContainerTerms,
            pdf_id: int,
            page_num: int,
            candidate: Term,
        ) -> None:
            candidate_lemma = candidate.lemma()
            container_terms.container_terms[
                candidate_lemma
            ] = container_terms.container_terms.get(candidate_lemma, set())

            num_tokens = len(candidate.tokens)
            for i in range(num_tokens):
                jstart, jstop = i + 1, (num_tokens + 1 if i > 0 else num_tokens)
                for j in range(jstart, jstop):
                    subcandidate = Term(
                        candidate.tokens[i:j],
                        candidate.fontsize,
                        candidate.ncolor,
                        candidate.augmented,
                    )
                    subcandidate_lemma = subcandidate.lemma()
                    if subcandidate_lemma not in domain_candidates_set:
                        continue

                    container_term_set = container_terms.container_terms.get(
                        subcandidate_lemma, set()
                    )
                    container_term_set.add(candidate_lemma)
                    container_terms.container_terms[
                        subcandidate_lemma
                    ] = container_term_set

        container_terms = self._runner.run_through_candidates(
            domain_candidates,
            DomainContainerTerms(domain_candidates.domain, dict()),
            update,
        )
        return container_terms



import json
import os
from glob import glob
from shutil import rmtree
from typing import Union

from py_pdf_term.stylings import PDFStylingScoreList

from ...configs import StylingLayerConfig
from ..util import create_dir_name_from_config, create_file_name_from_path
from .base import BaseStylingLayerCache


class StylingLayerFileCache(BaseStylingLayerCache):
    def __init__(self, cache_dir: str) -> None:
        super().__init__(cache_dir)
        self._cache_dir = cache_dir

    def load(
        self, pdf_path: str, config: StylingLayerConfig
    ) -> Union[PDFStylingScoreList, None]:
        dir_name = create_dir_name_from_config(config)
        file_name = create_file_name_from_path(pdf_path, "json")
        cache_file_path = os.path.join(self._cache_dir, dir_name, file_name)

        if not os.path.isfile(cache_file_path):
            return None

        with open(cache_file_path, "r") as json_file:
            try:
                obj = json.load(json_file)
            except json.JSONDecodeError:
                return None

        return PDFStylingScoreList.from_dict(obj)

    def store(
        self, styling_scores: PDFStylingScoreList, config: StylingLayerConfig
    ) -> None:
        dir_name = create_dir_name_from_config(config)
        file_name = create_file_name_from_path(styling_scores.pdf_path, "json")
        cache_file_path = os.path.join(self._cache_dir, dir_name, file_name)

        os.makedirs(os.path.dirname(cache_file_path), exist_ok=True)

        with open(cache_file_path, "w") as json_file:
            json.dump(styling_scores.to_dict(), json_file, ensure_ascii=False, indent=2)

    def remove(self, pdf_path: str, config: StylingLayerConfig) -> None:
        dir_name = create_dir_name_from_config(config)
        file_name = create_file_name_from_path(pdf_path, "json")
        cache_dir_path = os.path.join(self._cache_dir, dir_name)
        cache_file_path = os.path.join(cache_dir_path, file_name)

        if not os.path.isfile(cache_file_path):
            return

        os.remove(cache_file_path)

        cache_file_paths = glob(os.path.join(cache_dir_path, "*.json"))
        if not cache_file_paths:
            rmtree(cache_dir_path)



import re
from typing import List, Optional

from py_pdf_term._common.consts import ALPHABET_REGEX, NUMBER_REGEX
from py_pdf_term._common.data import Term

from .base import BaseSplitter
from ..classifiers import BaseTokenClassifier


class SymbolNameSplitter(BaseSplitter):
    def __init__(self, classifiers: Optional[List[BaseTokenClassifier]] = None) -> None:
        super().__init__(classifiers=classifiers)

    def split(self, term: Term) -> List[Term]:
        num_tokens = len(term.tokens)
        if num_tokens < 2:
            return [term]

        regex = re.compile(rf"{ALPHABET_REGEX}|{NUMBER_REGEX}+|\-")
        last_str = str(term.tokens[len(term.tokens) - 1])
        second_last_str = str(term.tokens[len(term.tokens) - 2])

        if not regex.fullmatch(last_str) or regex.fullmatch(second_last_str):
            return [term]

        nonsym_tokens = term.tokens[: num_tokens - 1]
        nonsym_term = Term(nonsym_tokens, term.fontsize, term.ncolor, term.augmented)
        return [nonsym_term]



from abc import ABCMeta, abstractmethod
from typing import Generic

from py_pdf_term.candidates import DomainCandidateTermList

from ..rankingdata.base import RankingData


class BaseRankingDataCollector(Generic[RankingData], metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def collect(self, domain_candidates: DomainCandidateTermList) -> RankingData:
        raise NotImplementedError(f"{self.__class__.__name__}.collect()")



from dataclasses import dataclass, field
from typing import List

from .base import BaseLayerConfig


@dataclass(frozen=True)
class StylingLayerConfig(BaseLayerConfig):
    styling_scores: List[str] = field(
        default_factory=lambda: [
            "py_pdf_term.FontsizeScore",
            "py_pdf_term.ColorScore",
        ]
    )
    cache: str = "py_pdf_term.StylingLayerFileCache"



from py_pdf_term.techterms import PDFTechnicalTermList

from .data import DomainPDFList
from .extractor import PyPDFTermExtractor

# isort: unique-list
__all__ = ["DomainPDFList", "PDFTechnicalTermList", "PyPDFTermExtractor"]



from typing import Any, Type

from ...caches import (
    BaseMethodLayerDataCache,
    BaseMethodLayerRankingCache,
    MethodLayerDataFileCache,
    MethodLayerDataNoCache,
    MethodLayerRankingFileCache,
    MethodLayerRankingNoCache,
)
from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class MethodLayerRankingCacheMapper(BaseMapper[Type[BaseMethodLayerRankingCache]]):
    @classmethod
    def default_mapper(cls) -> "MethodLayerRankingCacheMapper":
        default_mapper = cls()

        cache_clses = [MethodLayerRankingNoCache, MethodLayerRankingFileCache]
        for cache_cls in cache_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{cache_cls.__name__}", cache_cls)

        return default_mapper


class MethodLayerDataCacheMapper(BaseMapper[Type[BaseMethodLayerDataCache[Any]]]):
    @classmethod
    def default_mapper(cls) -> "MethodLayerDataCacheMapper":
        default_mapper = cls()

        cache_clses = [
            ("MethodLayerDataNoCache", MethodLayerDataNoCache[Any]),
            ("MethodLayerDataFileCache", MethodLayerDataFileCache[Any]),
        ]
        for name, cache_cls in cache_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{name}", cache_cls)

        return default_mapper



# pyright:reportUnknownMemberType=false
# pyright:reportUnknownArgumentType=false
# pyright:reportUnknownLambdaType=false

import re
from itertools import accumulate
from typing import Any, List

import ja_core_news_sm

from py_pdf_term._common.consts import JAPANESE_REGEX, NOSPACE_REGEX, SYMBOL_REGEX

from .data import Token
from .base import BaseLanguageTokenizer

SPACES = re.compile(r"\s+")
DELIM_SPACE = re.compile(rf"(?<={NOSPACE_REGEX}) (?={NOSPACE_REGEX})")


class JapaneseTokenizer(BaseLanguageTokenizer):
    def __init__(self) -> None:
        enable_pipes = []
        self._model = ja_core_news_sm.load()
        self._model.select_pipes(enable=enable_pipes)

        self._ja_regex = re.compile(JAPANESE_REGEX)
        self._symbol_regex = re.compile(rf"({SYMBOL_REGEX})")

    def inscope(self, text: str) -> bool:
        return self._ja_regex.search(text) is not None

    def tokenize(self, text: str) -> List[Token]:
        text = SPACES.sub(" ", text).strip()
        orginal_space_pos = {
            match.start() - offset
            for offset, match in enumerate(re.finditer(r" ", text))
            if DELIM_SPACE.match(text, match.start()) is not None
        }

        text = DELIM_SPACE.sub("", text)
        text = self._symbol_regex.sub(r" \1 ", text)
        tokens = list(map(self._create_token, self._model(text)))

        if not orginal_space_pos:
            return tokens

        tokenized_space_pos = set(
            accumulate(map(lambda token: len(str(token)), tokens))
        )
        if not orginal_space_pos.issubset(tokenized_space_pos):
            return tokens

        pos, i = 0, 0
        num_token = len(tokens) + len(orginal_space_pos)
        while i < num_token:
            if pos in orginal_space_pos:
                pos += len(str(tokens[i]))
                space = Token("ja", " ", "空白", "*", "*", " ")
                tokens.insert(i, space)
                i += 2
            else:
                pos += len(str(tokens[i]))
                i += 1

        return tokens

    def _create_token(self, token: Any) -> Token:
        if self._symbol_regex.fullmatch(token.text):
            return Token("ja", token.text, "補助記号", "一般", "*", token.text)

        pos_with_categories = token.tag_.split("-")
        num_categories = len(pos_with_categories) - 1

        pos = pos_with_categories[0]
        category = pos_with_categories[1] if num_categories >= 1 else "*"
        subcategory = pos_with_categories[2] if num_categories >= 2 else "*"

        return Token("ja", token.text, pos, category, subcategory, token.lemma_.lower())



from py_pdf_term._common.data import ScoredTerm, Term
from py_pdf_term._common.extended_math import extended_log10
from py_pdf_term.candidates import DomainCandidateTermList

from ..data import MethodTermRanking
from ..rankingdata import MCValueRankingData
from .base import BaseSingleDomainRanker


class MCValueRanker(BaseSingleDomainRanker[MCValueRankingData]):
    def __init__(self) -> None:
        pass

    def rank_terms(
        self,
        domain_candidates: DomainCandidateTermList,
        ranking_data: MCValueRankingData,
    ) -> MethodTermRanking:
        domain_candidates_dict = domain_candidates.to_nostyle_candidates_dict(
            to_str=lambda candidate: candidate.lemma()
        )
        ranking = list(
            map(
                lambda candidate: self._calculate_score(candidate, ranking_data),
                domain_candidates_dict.values(),
            )
        )
        ranking.sort(key=lambda term: -term.score)
        return MethodTermRanking(domain_candidates.domain, ranking)

    def _calculate_score(
        self, candidate: Term, ranking_data: MCValueRankingData
    ) -> ScoredTerm:
        candidate_lemma = candidate.lemma()

        term_freq = ranking_data.term_freq.get(candidate_lemma, 0)
        container_terms = ranking_data.container_terms.get(candidate_lemma, set())
        num_containers = len(container_terms)
        container_freq = sum(
            map(
                lambda container: ranking_data.term_freq.get(container, 0),
                container_terms,
            )
        )

        term_len_score = extended_log10(len(candidate.tokens))
        freq_score = extended_log10(
            term_freq - container_freq / num_containers
            if num_containers > 0
            else term_freq
        )
        score = term_len_score + freq_score
        return ScoredTerm(candidate_lemma, score)



from typing import Any, Callable, Dict, List, Union

from py_pdf_term.methods import MethodTermRanking
from py_pdf_term.methods._methods.rankingdata import RankingData

from ...configs import MethodLayerConfig
from .base import BaseMethodLayerDataCache, BaseMethodLayerRankingCache


class MethodLayerRankingNoCache(BaseMethodLayerRankingCache):
    def __init__(self, cache_dir: str) -> None:
        pass

    def load(
        self,
        pdf_paths: List[str],
        config: MethodLayerConfig,
    ) -> Union[MethodTermRanking, None]:
        pass

    def store(
        self,
        pdf_paths: List[str],
        term_ranking: MethodTermRanking,
        config: MethodLayerConfig,
    ) -> None:
        pass

    def remove(self, pdf_paths: List[str], config: MethodLayerConfig) -> None:
        pass


class MethodLayerDataNoCache(BaseMethodLayerDataCache[RankingData]):
    def __init__(self, cache_dir: str) -> None:
        pass

    def load(
        self,
        pdf_paths: List[str],
        config: MethodLayerConfig,
        from_dict: Callable[[Dict[str, Any]], RankingData],
    ) -> Union[RankingData, None]:
        pass

    def store(
        self,
        pdf_paths: List[str],
        ranking_data: RankingData,
        config: MethodLayerConfig,
    ) -> None:
        pass

    def remove(self, pdf_paths: List[str], config: MethodLayerConfig) -> None:
        pass



from typing import Any, Type

from py_pdf_term.methods import (
    BaseSingleDomainRankingMethod,
    FLRHMethod,
    FLRMethod,
    HITSMethod,
    MCValueMethod,
)

from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class SingleDomainRankingMethodMapper(
    BaseMapper[Type[BaseSingleDomainRankingMethod[Any]]]
):
    @classmethod
    def default_mapper(cls) -> "SingleDomainRankingMethodMapper":
        default_mapper = cls()

        single_domain_clses = [MCValueMethod, FLRMethod, HITSMethod, FLRHMethod]
        for method_cls in single_domain_clses:
            default_mapper.add(f"{PACKAGE_NAME}.{method_cls.__name__}", method_cls)

        return default_mapper



from typing import Dict, List, Optional

from .data import ScoredTerm


def ranking_to_dict(
    ranking: List[ScoredTerm], rate: Optional[float] = None
) -> Dict[str, float]:
    if rate is None:
        return {item.term: item.score for item in ranking}

    threshold = ranking[int(rate * len(ranking))].score
    return {item.term: item.score for item in ranking if item.score > threshold}



from typing import List, Optional

from .data import Token
from .base import BaseLanguageTokenizer
from .english import EnglishTokenizer
from .japanese import JapaneseTokenizer


class Tokenizer:
    def __init__(
        self, lang_tokenizers: Optional[List[BaseLanguageTokenizer]] = None
    ) -> None:
        if lang_tokenizers is None:
            lang_tokenizers = [JapaneseTokenizer(), EnglishTokenizer()]

        self._lang_tokenizers = lang_tokenizers

    def tokenize(self, text: str) -> List[Token]:
        if not text:
            return []

        for tokenizer in self._lang_tokenizers:
            if tokenizer.inscope(text):
                return tokenizer.tokenize(text)

        return []



import re
from xml.etree.ElementTree import Element

FLOAT_REGEX = r"\d+(?:\.\d+)?"
BRACKET_MONO_REGEX = rf"\[({FLOAT_REGEX})\]"
PAREN_MONO_REGEX = rf"\(({FLOAT_REGEX})\)"
BRACKET_TRI_REGEX = rf"\[({FLOAT_REGEX})\s*,\s*({FLOAT_REGEX})\s*,\s*({FLOAT_REGEX})\]"
PAREN_TRI_REGEX = rf"\(({FLOAT_REGEX})\s*,\s*({FLOAT_REGEX})\s*,\s*({FLOAT_REGEX})\)"


def textnode_text(textnode: Element, default: str = "") -> str:
    return textnode.text or default


def textnode_fontsize(textnode: Element, default: float = 0.0) -> float:
    try:
        return float(textnode.get("size", default))
    except ValueError:
        return default


def textnode_ncolor(textnode: Element, default: str = str((0.0, 0.0, 0.0))) -> str:
    raw_ncolor = textnode.get("ncolor")
    if raw_ncolor is None:
        return default

    match = re.fullmatch(FLOAT_REGEX, raw_ncolor)
    if match:
        color = float(match[0])
        return str((color, color, color))

    match = re.fullmatch(BRACKET_MONO_REGEX, raw_ncolor)
    if match:
        color = float(match[1])
        return str((color, color, color))

    match = re.fullmatch(PAREN_MONO_REGEX, raw_ncolor)
    if match:
        color = float(match[1])
        return str((color, color, color))

    match = re.fullmatch(BRACKET_TRI_REGEX, raw_ncolor)
    if match:
        color = float(match[1]), float(match[2]), float(match[3])
        return str(color)

    match = re.fullmatch(PAREN_TRI_REGEX, raw_ncolor)
    if match:
        color = float(match[1]), float(match[2]), float(match[3])
        return str(color)

    return default



from typing import Optional

from py_pdf_term.pdftoxml import PDFnXMLElement, PDFtoXMLConverter

from ..caches import DEFAULT_CACHE_DIR
from ..configs import XMLLayerConfig
from ..mappers import BinaryOpenerMapper, XMLLayerCacheMapper


class XMLLayer:
    def __init__(
        self,
        config: Optional[XMLLayerConfig] = None,
        bin_opener_mapper: Optional[BinaryOpenerMapper] = None,
        cache_mapper: Optional[XMLLayerCacheMapper] = None,
        cache_dir: str = DEFAULT_CACHE_DIR,
    ) -> None:
        if config is None:
            config = XMLLayerConfig()
        if bin_opener_mapper is None:
            bin_opener_mapper = BinaryOpenerMapper.default_mapper()
        if cache_mapper is None:
            cache_mapper = XMLLayerCacheMapper.default_mapper()

        open_bin = bin_opener_mapper.find(config.open_bin)
        cache_cls = cache_mapper.find(config.cache)

        self._converter = PDFtoXMLConverter(open_bin=open_bin)
        self._cache = cache_cls(cache_dir=cache_dir)
        self._config = config

    def create_pdfnxml(self, pdf_path: str) -> PDFnXMLElement:
        pdfnxml = None
        pdfnxml = self._cache.load(pdf_path, self._config)

        if pdfnxml is None:
            pdfnxml = self._converter.convert_as_element(
                pdf_path,
                nfc_norm=self._config.nfc_norm,
                include_pattern=self._config.include_pattern,
                exclude_pattern=self._config.exclude_pattern,
            )

        self._cache.store(pdfnxml, self._config)

        return pdfnxml

    def remove_cache(self, pdf_path: str) -> None:
        self._cache.remove(pdf_path, self._config)



from abc import ABCMeta, abstractmethod
from typing import Union

from py_pdf_term.pdftoxml import PDFnXMLElement

from ...configs import XMLLayerConfig


class BaseXMLLayerCache(metaclass=ABCMeta):
    def __init__(self, cache_dir: str) -> None:
        pass

    @abstractmethod
    def load(
        self, pdf_path: str, config: XMLLayerConfig
    ) -> Union[PDFnXMLElement, None]:
        raise NotImplementedError(f"{self.__class__.__name__}.load()")

    @abstractmethod
    def store(self, pdfnxml: PDFnXMLElement, config: XMLLayerConfig) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.store()")

    @abstractmethod
    def remove(self, pdf_path: str, config: XMLLayerConfig) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.remove()")



from typing import Dict, Optional

from py_pdf_term._common.data import ScoredTerm
from py_pdf_term.candidates import (
    DomainCandidateTermList,
    PageCandidateTermList,
    PDFCandidateTermList,
)
from py_pdf_term.methods import MethodTermRanking
from py_pdf_term.stylings import (
    DomainStylingScoreList,
    PageStylingScoreList,
    PDFStylingScoreList,
)

from .data import DomainTechnicalTermList, PageTechnicalTermList, PDFTechnicalTermList
from .utils import ranking_to_dict


class TechnicalTermExtractor:
    def __init__(self, max_num_terms: int = 10, acceptance_rate: float = 0.75) -> None:
        self._max_num_terms = max_num_terms
        self._acceptance_rate = acceptance_rate
        self._cache: Optional[Dict[str, float]] = None

    def extract_from_domain(
        self,
        domain_candidates: DomainCandidateTermList,
        term_ranking: MethodTermRanking,
        domain_styling_scores: DomainStylingScoreList,
    ) -> DomainTechnicalTermList:
        cache_should_flush = self._cache is None
        if self._cache is None:
            self._cache = ranking_to_dict(term_ranking.ranking, self._acceptance_rate)

        pdf_techterms = [
            self.extract_from_pdf(pdf_candidates, term_ranking, pdf_styling_scores)
            for pdf_candidates, pdf_styling_scores in zip(
                domain_candidates.pdfs, domain_styling_scores.pdfs
            )
        ]

        if cache_should_flush:
            self._cache = None

        return DomainTechnicalTermList(domain_candidates.domain, pdf_techterms)

    def extract_from_pdf(
        self,
        pdf_candidates: PDFCandidateTermList,
        term_ranking: MethodTermRanking,
        pdf_styling_scores: PDFStylingScoreList,
    ) -> PDFTechnicalTermList:
        cache_should_flush = self._cache is None
        if self._cache is None:
            self._cache = ranking_to_dict(term_ranking.ranking, self._acceptance_rate)

        page_techterms = [
            self._extract_from_page(page_candidates, term_ranking, page_styling_scores)
            for page_candidates, page_styling_scores in zip(
                pdf_candidates.pages, pdf_styling_scores.pages
            )
        ]

        if cache_should_flush:
            self._cache = None

        return PDFTechnicalTermList(pdf_candidates.pdf_path, page_techterms)

    def _extract_from_page(
        self,
        page_candidates: PageCandidateTermList,
        term_ranking: MethodTermRanking,
        page_styling_scores: PageStylingScoreList,
    ) -> PageTechnicalTermList:
        cache_should_flush = self._cache is None
        if self._cache is None:
            self._cache = ranking_to_dict(term_ranking.ranking, self._acceptance_rate)

        method_score_dict = self._cache
        styling_score_dict = ranking_to_dict(page_styling_scores.ranking)

        def term_score(term_lemma: str) -> float:
            method_score = method_score_dict[term_lemma]
            styling_score = styling_score_dict[term_lemma]
            if method_score >= 0.0:
                return method_score * styling_score
            else:
                return method_score / styling_score

        scored_terms = [
            ScoredTerm(term_str, term_score(term.lemma()))
            for term_str, term in page_candidates.to_nostyle_candidates_dict().items()
            if term.lemma() in method_score_dict and term.lemma() in styling_score_dict
        ]

        if len(scored_terms) > self._max_num_terms:
            scores = list(map(lambda scored_term: scored_term.score, scored_terms))
            scores.sort(reverse=True)
            threshold = scores[self._max_num_terms]
            scored_terms = list(
                filter(lambda scored_term: scored_term.score > threshold, scored_terms)
            )

        if cache_should_flush:
            self._cache = None

        return PageTechnicalTermList(page_candidates.page_num, scored_terms)



from typing import List, Optional

from py_pdf_term.candidates import (
    CandidateTermExtractor,
    DomainCandidateTermList,
    PDFCandidateTermList,
)

from ..caches import DEFAULT_CACHE_DIR
from ..configs import CandidateLayerConfig
from ..data import DomainPDFList
from ..mappers import (
    AugmenterMapper,
    CandidateLayerCacheMapper,
    CandidateTermFilterMapper,
    CandidateTokenFilterMapper,
    LanguageTokenizerMapper,
    SplitterMapper,
    TokenClassifilerMapper,
)
from .xml import XMLLayer


class CandidateLayer:
    def __init__(
        self,
        xml_layer: XMLLayer,
        config: Optional[CandidateLayerConfig] = None,
        lang_tokenizer_mapper: Optional[LanguageTokenizerMapper] = None,
        token_classifier_mapper: Optional[TokenClassifilerMapper] = None,
        token_filter_mapper: Optional[CandidateTokenFilterMapper] = None,
        term_filter_mapper: Optional[CandidateTermFilterMapper] = None,
        splitter_mapper: Optional[SplitterMapper] = None,
        augmenter_mapper: Optional[AugmenterMapper] = None,
        cache_mapper: Optional[CandidateLayerCacheMapper] = None,
        cache_dir: str = DEFAULT_CACHE_DIR,
    ) -> None:
        if config is None:
            config = CandidateLayerConfig()
        if lang_tokenizer_mapper is None:
            lang_tokenizer_mapper = LanguageTokenizerMapper.default_mapper()
        if token_classifier_mapper is None:
            token_classifier_mapper = TokenClassifilerMapper.default_mapper()
        if token_filter_mapper is None:
            token_filter_mapper = CandidateTokenFilterMapper.default_mapper()
        if term_filter_mapper is None:
            term_filter_mapper = CandidateTermFilterMapper.default_mapper()
        if splitter_mapper is None:
            splitter_mapper = SplitterMapper.default_mapper()
        if augmenter_mapper is None:
            augmenter_mapper = AugmenterMapper.default_mapper()
        if cache_mapper is None:
            cache_mapper = CandidateLayerCacheMapper.default_mapper()

        lang_tokenizer_clses = lang_tokenizer_mapper.bulk_find(config.lang_tokenizers)
        classifier_clses = token_classifier_mapper.bulk_find(config.token_classifiers)
        token_filter_clses = token_filter_mapper.bulk_find(config.token_filters)
        term_filter_clses = term_filter_mapper.bulk_find(config.term_filters)
        splitter_clses = splitter_mapper.bulk_find(config.splitters)
        augmenter_clses = augmenter_mapper.bulk_find(config.augmenters)
        cache_cls = cache_mapper.find(config.cache)

        self._extractor = CandidateTermExtractor(
            lang_tokenizer_clses=lang_tokenizer_clses,
            token_classifier_clses=classifier_clses,
            token_filter_clses=token_filter_clses,
            term_filter_clses=term_filter_clses,
            splitter_clses=splitter_clses,
            augmenter_clses=augmenter_clses,
        )
        self._cache = cache_cls(cache_dir=cache_dir)
        self._config = config

        self._xml_layer = xml_layer

    def create_domain_candiates(
        self, domain_pdfs: DomainPDFList
    ) -> DomainCandidateTermList:
        pdf_candidates_list: List[PDFCandidateTermList] = []
        for pdf_path in domain_pdfs.pdf_paths:
            pdf_candidates = self.create_pdf_candidates(pdf_path)
            pdf_candidates_list.append(pdf_candidates)

        return DomainCandidateTermList(domain_pdfs.domain, pdf_candidates_list)

    def create_pdf_candidates(self, pdf_path: str) -> PDFCandidateTermList:
        pdf_candidates = self._cache.load(pdf_path, self._config)

        if pdf_candidates is None:
            pdfnxml = self._xml_layer.create_pdfnxml(pdf_path)
            pdf_candidates = self._extractor.extract_from_xml_element(pdfnxml)

        self._cache.store(pdf_candidates, self._config)

        return pdf_candidates

    def remove_cache(self, pdf_path: str) -> None:
        self._cache.remove(pdf_path, self._config)



from py_pdf_term._common.data import ScoredTerm, Term
from py_pdf_term._common.extended_math import extended_log10
from py_pdf_term.candidates import DomainCandidateTermList

from ..data import MethodTermRanking
from ..rankingdata import FLRRankingData
from .base import BaseSingleDomainRanker


class FLRRanker(BaseSingleDomainRanker[FLRRankingData]):
    def __init__(self) -> None:
        pass

    def rank_terms(
        self, domain_candidates: DomainCandidateTermList, ranking_data: FLRRankingData
    ) -> MethodTermRanking:
        domain_candidates_dict = domain_candidates.to_nostyle_candidates_dict(
            to_str=lambda candidate: candidate.lemma()
        )
        ranking = list(
            map(
                lambda candidate: self._calculate_score(candidate, ranking_data),
                domain_candidates_dict.values(),
            )
        )
        ranking.sort(key=lambda term: -term.score)
        return MethodTermRanking(domain_candidates.domain, ranking)

    def _calculate_score(
        self, candidate: Term, ranking_data: FLRRankingData
    ) -> ScoredTerm:
        candidate_lemma = candidate.lemma()
        num_tokens = len(candidate.tokens)
        num_meaningless_tokens = sum(
            map(lambda token: 1 if token.is_meaningless else 0, candidate.tokens)
        )
        term_freq_score = extended_log10(ranking_data.term_freq.get(candidate_lemma, 0))

        concat_score = 0.0
        for token in candidate.tokens:
            if token.is_meaningless:
                continue

            lscore = sum(ranking_data.left_freq.get(token.lemma, dict()).values())
            rscore = sum(ranking_data.right_freq.get(token.lemma, dict()).values())
            concat_score += 0.5 * (extended_log10(lscore) + extended_log10(rscore))

        concat_score /= num_tokens - num_meaningless_tokens

        score = term_freq_score + concat_score
        return ScoredTerm(candidate_lemma, score)



from .base import BaseMethodLayerDataCache, BaseMethodLayerRankingCache
from .file import MethodLayerDataFileCache, MethodLayerRankingFileCache
from .nocache import MethodLayerDataNoCache, MethodLayerRankingNoCache

# isort: unique-list
__all__ = [
    "BaseMethodLayerDataCache",
    "BaseMethodLayerRankingCache",
    "MethodLayerDataFileCache",
    "MethodLayerDataNoCache",
    "MethodLayerRankingFileCache",
    "MethodLayerRankingNoCache",
]



from typing import Type

from py_pdf_term.stylings.scores import BaseStylingScore, ColorScore, FontsizeScore

from ..base import BaseMapper
from ..consts import PACKAGE_NAME


class StylingScoreMapper(BaseMapper[Type[BaseStylingScore]]):
    @classmethod
    def default_mapper(cls) -> "StylingScoreMapper":
        default_mapper = cls()

        styling_score_clses = [FontsizeScore, ColorScore]
        for styling_score_cls in styling_score_clses:
            default_mapper.add(
                f"{PACKAGE_NAME}.{styling_score_cls.__name__}", styling_score_cls
            )

        return default_mapper



from dataclasses import dataclass

from .base import BaseLayerConfig


@dataclass(frozen=True)
class TechnicalTermLayerConfig(BaseLayerConfig):
    max_num_terms: int = 10
    acceptance_rate: float = 0.75
